#
# Utils to help out yaml dump
#


# Provide better formatting of multiline literals
# from: https://github.com/cloudlinux/kuberdock-platform/blob/master/kubedock/kapi/apps.py
def str_presenter(dumper, data):
  # check for multiline strings
  if len(data.splitlines()) == 1 and data[-1] == '\n':
    return dumper.represent_scalar(
      'tag:yaml.org,2002:str', data, style='>')
  if len(data.splitlines()) > 1:
    return dumper.represent_scalar(
      'tag:yaml.org,2002:str', data, style='|')
  return dumper.represent_scalar(
    'tag:yaml.org,2002:str', data.strip())


from janis import String, ToolInput, ToolOutput, Stdout
from janis.unix.tools.unixtool import UnixTool


class Echo(UnixTool):

    @staticmethod
    def tool():
        return "Echo"

    def friendly_name(self):
        return "Echo - Print to console"

    @staticmethod
    def base_command():
        return "echo"

    def inputs(self):
        return [
            ToolInput("input", String())
        ]

    def outputs(self):
        return [ToolOutput("output", Stdout())]



from janis.types.common_data_types import File


class TarFile(File):
    @staticmethod
    def name():
        return "TarFile"

    def doc(self):
        return "A tarfile, ending with .tar"


# tar_file = InputType('tar_file', label='a unix tar archive', description='A unix compressed or uncompressed tar archive')
#
#
# class TarFileFactory(InputFactory):
#   @classmethod
#   def type(cls) -> InputType:
#     return tar_file
#
#   @classmethod
#   def schema(cls):
#     return {
#       'schema': {
#         'path': {'type': 'string'},
#         'label': {'type': 'string'}
#       },
#       'nullable': True
#     }
#
#   @classmethod
#   def build(cls, label: str, input_dict: dict, debug=False):
#     return TarFile(label, input_dict, debug=debug)
#
#
# class TarFile(Input):
#   def translate_for_workflow(self) -> Dict[str, str]:
#     return {self.id(): 'File'}
#
#   def translate_for_input(self):
#     fd = {'class': 'File', 'path': self._path}
#     return {self.id(): fd}
#
#   def resolve(self):
#     pass
#
#   def __init__(self, label: str, input_dict: dict, debug=False):
#     super().__init__(label, input_dict)
#     self._resolved = True
#     self._debug = debug
#     self._path = [self.meta()['path']]
#     self._resolved = False
#
#   def identify(self):
#     super().identify()
#     if self._debug:
#       print("Path:", self._path)
#
#   def datum_type(self):
#     return self.type()
#
#   def is_subtype_of(self, other):
#     return False




#
# Untar a file

from janis import Array
from janis.tool.commandtool import ToolInput, ToolOutput
from janis.types.common_data_types import File, Filename
from janis.unix.tools.unixtool import UnixTool


class Tar(UnixTool):

    @staticmethod
    def tool():
        return "Tar"

    def friendly_name(self):
        return "Tar archive"

    @staticmethod
    def base_command():
        return ["tar", "cvf"]

    def inputs(self):
        return [
            ToolInput("files", Array(File()), position=2),
            ToolInput("files2", Array(File(), optional=True), position=3),
            ToolInput("tarName", Filename(extension=".tar"), position=1)
        ]

    def outputs(self):
        return [
            ToolOutput("tarred", File(), glob="$(inputs.tarName)")
        ]


if __name__ == "__main__":
    print(Tar().help())

"""

EXAMPLE XML

        <Step>
            <title>Tar</title>
            <author>Michael Franklin</author>
            <cores>1</cores>
            <ram>1024</ram>
            <label>{{label}}</label>
            <supportedTypes>
                <type>cwl</type>
                <type>wdl</type>
            </supportedTypes>
        
            <inputs>
                <input>
                    <tag>input2</tag>
                    <type>File</type>
                </input>
                <input>
                    <tag>input2</tag>
                    <type>File</type>
                </input>
            </inputs>
            <outputs>
                <output>
                    <tag>output</tag>
                    <type>File</type>
                </output>
            </outputs>
        </Step>

==============================

        class: CommandLineTool
        cwlVersion: v1.0
        id: tar
        baseCommand:
          - tar
          - cvf
        inputs:
          - id: tarName
            type: string?
            inputBinding:
              position: 0
          - id: input1
            type:
              - File
            inputBinding:
              position: 1
          - id: input2
            type:
              - File
            inputBinding:
              position: 2
        outputs:
          - id: output
            type: File?
            outputBinding:
              glob: '*'
        label: tar

  ========================================
  
        task SingleTar {
            String tarName
            File input1
            File input2
        
            command {
                tar cvf ${tarName} ${input1} ${input2}
            }
        
            output {
                File out = glob("*")[0]
            }
        }
"""


from abc import ABC, abstractmethod
from typing import Optional, List, Dict, Any
import re

from janis.types.common_data_types import Array
from janis.utils.logger import Logger
from janis.types.data_types import DataType, NativeTypes
# import cwlgen.cwlgen as cwl
from janis.utils.metadata import Metadata

ToolType = str


class ToolTypes:
    Workflow: ToolType = "workflow"
    CommandTool: ToolType = "command-tool"
    ExpressionTool: ToolType = "expression-tool"


class ToolArgument:
    expr_pattern = "\$\(.*\)"

    def __init__(self, value: Any, prefix: Optional[str] = None, position: Optional[int] = 0,
                 separate_value_from_prefix=None, doc: Optional[str]=None, shell_quote: bool=None):
        self.prefix: Optional[str] = prefix
        self.value = value
        self.position: Optional[int] = position
        self.is_expression = (re.match(self.expr_pattern, self.value) is not None) if self.value else None
        self.separate_value_from_prefix = separate_value_from_prefix
        self.doc = doc
        self.shell_quote = shell_quote

        if self.prefix and self.separate_value_from_prefix is not None \
                and not self.separate_value_from_prefix and not self.prefix.endswith("="):
            # I don't really know what this means.
            Logger.warn(f"Argument ({self.prefix} {self.value}) is not separating and did not end with ='")


class ToolInput(ToolArgument):
    def __init__(self, tag: str, input_type: DataType, position: Optional[int] = None, prefix: Optional[str] = None,
                 separate_value_from_prefix: bool = None, default: Any = None, doc: Optional[str]=None,
                 nest_input_binding_on_array: bool=True, shell_quote=None):
        """
        :param tag: tag for input, what the yml will reference (eg: input1: path/to/file)
        :param input_type:
        """
        super().__init__(value=None, prefix=prefix, position=position,
                         separate_value_from_prefix=separate_value_from_prefix, doc=doc, shell_quote=shell_quote)

        if default is not None:
            input_type.optional = True

        self.tag: str = tag
        self.input_type: DataType = input_type
        self.optional = self.input_type.optional
        self.default = default
        self.nest_input_binding_on_array = nest_input_binding_on_array

    # def cwl(self):
    #     import cwlgen
    #     default = self.default if self.default else self.input_type.default()
    #
    #     data_type = self.input_type.cwl_type()
    #     input_binding = cwlgen.CommandLineBinding(
    #         # load_contents=self.load_contents,
    #         position=self.position,
    #         prefix=self.prefix,
    #         separate=self.separate_value_from_prefix,
    #         # item_separator=self.item_separator,
    #         # value_from=self.value_from,
    #         shell_quote=self.shell_quote,
    #     )
    #
    #     # Binding array inputs onto the console
    #     # https://www.commonwl.org/user_guide/09-array-inputs/
    #     if isinstance(self.input_type, Array) and isinstance(data_type, cwlgen.CommandInputArraySchema):
    #         if self.nest_input_binding_on_array:
    #             input_binding.prefix = None
    #             input_binding.separate = None
    #             nested_binding = cwlgen.CommandLineBinding(
    #                 # load_contents=self.load_contents,
    #                 prefix=self.prefix,
    #                 separate=self.separate_value_from_prefix,
    #                 # item_separator=self.item_separator,
    #                 # value_from=self.value_from,
    #                 shell_quote=self.shell_quote,
    #             )
    #             data_type.inputBinding = nested_binding
    #         else:
    #             input_binding.itemSeparator = ","
    #
    #     return cwlgen.CommandInputParameter(
    #         param_id=self.tag,
    #         label=self.tag,
    #         secondary_files=self.input_type.secondary_files(),
    #         # streamable=None,
    #         doc=self.doc,
    #         input_binding=input_binding,
    #         default=default,
    #         param_type=data_type
    #     )

    def id(self):
        return self.tag


class ToolOutput:
    def __init__(self, tag: str, output_type: DataType, glob: Optional[str] = None, doc: Optional[str]=None):
        self.tag = tag
        self.output_type: DataType = output_type
        self.glob = glob
        self.doc = doc

    def id(self):
        return self.tag


class Tool(ABC, object):
    """
    One of Workflow, CommandLineTool, ExpressionTool* (* unimplemented)
    """

    @classmethod
    @abstractmethod
    def type(cls) -> ToolType:
        raise Exception(f"'{cls}' must implement type() method")

    @abstractmethod
    def id(self) -> str:
        raise Exception("Must implement id() method")

    @staticmethod
    def tool_module():
        return None

    @staticmethod
    def tool_provider():
        return None

    @abstractmethod
    def inputs(self) -> List[ToolInput]:
        raise Exception("Must implement inputs() method")

    @abstractmethod
    def outputs(self) -> List[ToolOutput]:
        raise Exception("Must implement outputs() method")

    def inputs_map(self) -> Dict[str, ToolInput]:
        return {inp.tag: inp for inp in self.inputs()}

    def outputs_map(self) -> Dict[str, ToolOutput]:
        return {outp.tag: outp for outp in self.outputs()}

    def wdl(self, with_docker=True):
        raise Exception("Must implement wdl() method")

    @abstractmethod
    def friendly_name(self) -> str:
        # maps to CWL label (still exploring for WDL)
        raise Exception("Tools must implement friendly_name() method")

    def metadata(self) -> Optional[Metadata]:
        return None

    @staticmethod
    def version():
        return None

    def doc(self) -> Optional[str]:
        return None

    def help(self):
        import inspect
        path = inspect.getfile(self.__class__)

        ins = sorted(self.inputs(), key=lambda i: (i.position if i.position is not None else 0))

        def input_format(t: ToolInput):
            prefix_with_space = ""
            if t.prefix is not None:
                prefix_with_space = (t.prefix + ": ") if t.separate_value_from_prefix else t.prefix
            return f"\t\t{t.tag} ({prefix_with_space}{t.input_type.id()}" \
                f"{('=' + str(t.default)) if t.default is not None else ''}): {'' if t.doc is None else t.doc}"

        output_format = lambda t: f"\t\t{t.tag} ({t.output_type.id()}): {'' if t.doc is None else t.doc}"

        requiredInputs = "\n".join(input_format(x) for x in ins if not x.optional)
        optionalInputs = "\n".join(input_format(x) for x in ins if x.optional)
        outputs = "\n".join(output_format(o) for o in self.outputs())

        meta = self.metadata() if self.metadata() else Metadata()

        fn = self.friendly_name() if self.friendly_name() else self.id()
        en = f" ({self.id()})" if fn != self.id() else ""

        return f"""
Pipeline tool: {path} ({fn})
NAME
    {fn + en}

DESCRIPTION
    {meta.documentation if meta.documentation else "No documentation provided"}

INPUTS:
REQUIRED:
{requiredInputs}

OPTIONAL:
{optionalInputs}

OUTPUTS:
{outputs}
    """




"""
    Custom Types::

    We want to define some standard interface that allows for the implementation of types.
    Types must be supported as Inputs and Outputs, and must be comparable, ie we can compare
    the output of one step to another step. We'll guarantee that

    We'll also may require separate interfaces for CWL and WDL implementations, we may not be
    able to genericise the interface enough to support an automatic conversion

    We are allowed to require that a type must register itself when loaded.

"""
from abc import ABC, abstractmethod
from typing import Any, List, Dict, Optional
from wdlgen import WdlType

import cwlgen as cwl

NativeType = str


class NativeTypes:
    kStr: NativeType = "str"
    kInt: NativeType = "int"
    kLong: NativeType = "long"
    kFloat: NativeType = "float"
    kBool: NativeType = "bool"
    kDouble: NativeType = "double"
    kFile: NativeType = "file"
    kDirectory: NativeType = "dir"
    kArray: NativeType = "array"
    kStdout: NativeType = "stdout"

    _primitives: List[NativeType] = [kStr, kInt, kFloat, kLong, kDouble, kBool, kDouble]
    all: List[NativeType] = _primitives + [kFile, kDirectory, kArray]

    @staticmethod
    def is_primitive(t: NativeType) -> bool:
        return t in NativeTypes._primitives

    @staticmethod
    def is_valid(t: str):
        return t in NativeTypes.all

    @staticmethod
    def map_to_cwl(t: NativeType):
        if t == NativeTypes.kBool:
            return cwl.CwlTypes.BOOLEAN
        elif t == NativeTypes.kInt:
            return cwl.CwlTypes.INT
        elif t == NativeTypes.kLong:
            return cwl.CwlTypes.LONG
        elif t == NativeTypes.kFloat:
            return cwl.CwlTypes.FLOAT
        elif t == NativeTypes.kDouble:
            return cwl.CwlTypes.DOUBLE
        elif t == NativeTypes.kStr:
            return cwl.CwlTypes.STRING
        elif t == NativeTypes.kFile:
            return cwl.CwlTypes.FILE
        elif t == NativeTypes.kDirectory:
            return cwl.CwlTypes.DIRECTORY
        elif t == NativeTypes.kArray:
            return cwl.CwlTypes.ARRAY
        elif t == NativeTypes.kStdout:
            return cwl.CwlTypes.STDOUT
        raise Exception(f"Unhandled primitive type {t}, expected one of {', '.join(NativeTypes.all)}")

    @staticmethod
    def map_to_wdl(t: NativeType):
        if t == NativeTypes.kBool:
            return wdl.PrimitiveType.kBoolean
        elif t == NativeTypes.kInt:
            return wdl.PrimitiveType.kInt

        elif t == NativeTypes.kLong or t == NativeTypes.kFloat or t == NativeTypes.kDouble:
            return wdl.PrimitiveType.kFloat
        elif t == NativeTypes.kStr:
            return wdl.PrimitiveType.kString
        elif t == NativeTypes.kFile:
            return wdl.PrimitiveType.kFile
        elif t == NativeTypes.kStdout:
            return wdl.PrimitiveType.kFile
        elif t == NativeTypes.kDirectory:
            return None # wdl.PrimitiveType
        elif t == NativeTypes.kArray:
            return wdl.ArrayType.kArray
        raise Exception(f"Unhandled primitive type {t}, expected one of {', '.join(NativeTypes.all)}")

    @staticmethod
    def default_value(t: NativeType):
        if t == NativeTypes.kBool:
            return True
        elif t == NativeTypes.kInt:
            return 0
        elif t == NativeTypes.kLong:
            return 0.0
        elif t == NativeTypes.kFloat:
            return 0.0
        elif t == NativeTypes.kDouble:
            return 0.0
        elif t == NativeTypes.kStr:
            return "nothing"
        elif t == NativeTypes.kFile:
            return {"type": "File", "path": "path/to/file"}
        elif t == NativeTypes.kDirectory:
            return { "type": "Directory", "path": "path/to/file"}
        elif t == NativeTypes.kArray:
            return []


class DataType(ABC):

    def __init__(self, optional=False, default=None):
        self.optional = optional
        self.is_prim = NativeTypes.is_primitive(self.primitive())
        self.default_value = default

    @staticmethod
    @abstractmethod
    def name() -> str:
        raise Exception("Subclass MUST override name field")

    @staticmethod
    def secondary_files() -> Optional[List[str]]:
        return None

    @staticmethod
    @abstractmethod
    def primitive() -> NativeType:
        raise Exception("Subclass MUST override the 'primitive' method")

    @staticmethod
    @abstractmethod
    def doc() -> str:
        """
        Subclasses should override this class to provide additional information on how to
        correctly provide data to the class, what inputs it may have and what other types
        are compatible
        """
        raise Exception("Subclass MUST override the 'doc' field")

    @staticmethod
    def validate(meta: Any) -> bool:
        return True

    @classmethod
    # @abstractmethod
    def schema(cls) -> Dict:
        raise Exception("Subclass MUST override the 'schema' method")

    def get_value_from_meta(self, meta):
        return meta

    # The following methods don't need to be overriden, but can be

    def id(self):
        if self.optional:
            return f"Optional<{self.name()}>"
        return self.name()

    def identify(self):
        print(self.id())

    def can_receive_from(self, other) -> bool:
        """
        Can this class receive from $other, likely going to be type(a) == type(b)
        :param other:
        :return:
        """
        if not isinstance(other.received_type(), type(self.received_type())):
            return False
        if self.optional:
            # If I'm optional I can receive from optional / non optional
            return True
        # If I'm not optional, I must receive from not optional
        return not other.optional

    def received_type(self):
        """
        The type that will be received if joined from this type
        :return: mostly self, except for STDOUT | STDERR | STDIN
        """
        return self

    def input_field_from_input(self, meta):
        """
        Method to convert the field definition into a generic CWL-esque response
        :param meta:
        :return:
        """
        return None

    def _question_mark_if_optional(self):
        return "?" if self.optional else ""

    def cwl_type(self):
        return NativeTypes.map_to_cwl(self.primitive()) + self._question_mark_if_optional()

    def map_cwl_type(self, parameter: cwl.Parameter) -> cwl.Parameter:
        if not NativeTypes.is_valid(self.primitive()):
            raise Exception(f"{self.id()} must declare its primitive as one of the NativeTypes "
                            f"({', '.join(NativeTypes.all)})")

        parameter.type = NativeTypes.map_to_cwl(self.primitive()) + self._question_mark_if_optional()
        parameter.secondaryFiles = self.secondary_files()
        return parameter

    @staticmethod
    def cwl_input(value: Any):
        return value

    def wdl(self) -> WdlType:
        return WdlType.parse_type(NativeTypes.map_to_wdl(self.primitive()) + self._question_mark_if_optional())

    def default(self):
        return self.default_value




import os
import ruamel.yaml
import cwlgen
from typing import List, Dict, Type

from janis.hints import Hint, HintEnum, HintArray, HINTS
from janis.tool.tool import Tool, ToolInput
from janis.tool.commandtool import CommandTool
from janis.types.common_data_types import Int, Stdout, Array
from janis.utils.logger import Logger
from janis.utils.metadata import WorkflowMetadata, ToolMetadata
from janis.utils.janisconstants import RESOURCE_OVERRIDE_KEY, HINTS_KEY
from janis.workflow.step import StepNode


def dump_cwl(workflow, to_console=True, to_disk=False, with_docker=False, with_hints=False,
             with_resource_overrides=False, write_inputs_file=False):
    wf_cwl, inp_data, tools_cwl = translate_workflow(workflow,
                                                     with_docker=with_docker,
                                                     with_hints=with_hints,
                                                     with_resource_overrides=with_resource_overrides)

    wf_data = wf_cwl.get_dict()
    tools_data = [t.get_dict() for t in tools_cwl]

    ruamel.yaml.add_representer(cwlgen.utils.literal, cwlgen.utils.literal_presenter)

    if to_console:
        print(ruamel.yaml.dump(wf_data, default_flow_style=False))
        print(ruamel.yaml.dump(inp_data, default_flow_style=False))
        for t in tools_data:
            print(ruamel.yaml.dump(t, default_flow_style=False))

    if to_disk:
        d = os.path.expanduser("~") + f"/Desktop/{workflow.id()}/cwl/"
        d_tools = d + "tools/"

        if not os.path.isdir(d):
            os.makedirs(d)
        if not os.path.isdir(d_tools):
            os.makedirs(d_tools)

        os.chdir(d)
        wf_filename = d + workflow.id() + ".cwl"
        with open(wf_filename, "w+") as cwl:
            Logger.log(f"Writing {workflow.id()}.cwl to disk")
            ruamel.yaml.dump(wf_data, cwl, default_flow_style=False)
            Logger.log(f"Written {workflow.id()}.cwl to disk")

        if write_inputs_file:
            with open(d + workflow.id() + "-job.yml", "w+") as cwl:
                Logger.log(f"Writing {workflow.id()}-job.yml to disk")
                ruamel.yaml.dump(inp_data, cwl, default_flow_style=False)
                Logger.log(f"Written {workflow.id()}-job.yml to disk")
        else:
            Logger.log("Skipping writing input (yaml) job file")

        # z = zipfile.ZipFile(d + "tools.zip", "w")
        for tool in tools_data:
            tool_name = tool["id"]
            tool_filename = tool_name + ".cwl"
            with open(d_tools + tool_filename, "w+") as cwl:
                Logger.log(f"Writing {tool_name}.cwl to disk")
                ruamel.yaml.dump(tool, cwl, default_flow_style=False)
                Logger.log(f"Written {tool_name}.cwl to disk")

        import subprocess

        Logger.info("Validing outputted CWL")

        cwltool_result = subprocess.run(["cwltool", "--validate", wf_filename])
        if cwltool_result.returncode == 0:
            Logger.info("Exported workflow is valid CWL.")
        else:
            Logger.critical(cwltool_result.stderr)

        Logger.info("Zipping tools")
        os.chdir(d)

        zip_result = subprocess.run(["zip", "-r", "tools.zip", "tools/"])
        if zip_result.returncode == 0:
            Logger.info("Zipped tools")
        else:
            Logger.critical(zip_result.stderr)


def translate_workflow(wf, is_nested_tool=False, with_docker=False, with_hints=False,
                       with_resource_overrides=False):
    from janis.workflow.workflow import Workflow

    metadata = wf.metadata() if wf.metadata() else WorkflowMetadata()
    w = cwlgen.Workflow(wf.identifier, wf.friendly_name(), metadata.documentation)

    w.inputs: List[cwlgen.InputParameter] = [translate_input(i.input) for i in wf._inputs]

    if with_resource_overrides:
        rOverride = cwlgen.InputParameter(
            RESOURCE_OVERRIDE_KEY,
            param_type=["null", generate_cwl_resource_override_schema_for_steps(wf)]
        )
        w.inputs.append(rOverride)

    w.steps: List[cwlgen.WorkflowStep] = [translate_step(s, is_nested_tool=is_nested_tool) for s in wf._steps]

    w.outputs = [translate_output_node(o) for o in wf._outputs]

    #
    keys = ["coresMin", "coresMax", "ramMin", "ramMax"]
    sins = [translate_tool_input(ToolInput(k, Int(optional=True))) for k in keys]
    if with_resource_overrides:
        for s in w.steps:
            # work out whether (the tool of) s is a workflow or tool
            resource_override_step_inputs = [cwlgen.WorkflowStepInput(
                input_id=k,
                source=RESOURCE_OVERRIDE_KEY,
                value_from=f"${{var k = \"{k}\";var stepId = \"{s.id}\";if(!self) return null;if (!(stepId in self)) "
                            "return null;return self[stepId][k]}}"
            ) for k in keys]
            s.inputs.extend(resource_override_step_inputs)

    if with_hints:
        resource_schema = get_cwl_schema_for_recognised_hints()
        nullable_resource_schema = ["null", resource_schema]
        w.inputs.append(cwlgen.InputParameter(HINTS_KEY, param_type=nullable_resource_schema))
        for s in w.steps:
            s.inputs.append(cwlgen.WorkflowStepInput(HINTS_KEY, HINTS_KEY))

    w.requirements.append(cwlgen.InlineJavascriptReq())
    w.requirements.append(cwlgen.StepInputExpressionRequirement())

    if wf.has_scatter:
        w.requirements.append(cwlgen.ScatterFeatureRequirement())
    if wf.has_subworkflow:
        w.requirements.append(cwlgen.SubworkflowFeatureRequirement())
    if wf.has_multiple_inputs:
        w.requirements.append(cwlgen.MultipleInputFeatureRequirement())

    tools = []
    tools_to_build: Dict[str, Tool] = {s.step.tool().id(): s.step.tool() for s in wf._steps}
    for t in tools_to_build:
        tool: Tool = tools_to_build[t]
        if isinstance(tool, Workflow):
            wf_cwl, _, subtools = translate_workflow(tool,
                                                     is_nested_tool=True,
                                                     with_docker=with_docker,
                                                     with_hints=with_hints,
                                                     with_resource_overrides=with_resource_overrides)
            tools.append(wf_cwl)
            tools.extend(subtools)
        elif isinstance(tool, CommandTool):
            tool_cwl = translate_tool(tool, with_docker=with_docker)  # tool.cwl(with_docker=with_docker)
            if with_hints:
                tool_cwl.inputs.append(
                    cwlgen.InputParameter(HINTS_KEY, param_type=["null", get_cwl_schema_for_recognised_hints()]))
                hm = tool.hint_map()
                if hm:
                    tool_cwl.requirements.append(cwlgen.ResourceRequirement(
                        cores_min=generate_hint_selectors_for_hint_map("coresMin", hm),
                        cores_max=generate_hint_selectors_for_hint_map("coresMax", hm),
                        ram_min=generate_hint_selectors_for_hint_map("ramMin", hm),
                        ram_max=generate_hint_selectors_for_hint_map("ramMax", hm)
                    ))

            if with_resource_overrides:
                tool_cwl.inputs.extend(sins)

            tools.append(tool_cwl)
        else:
            raise Exception(f"Unknown tool type: '{type(tool)}'")

    inp = {i.id(): i.input.cwl_input() for i in wf._inputs}

    return w, inp, tools


def translate_tool(tool, with_docker):
    metadata = tool.metadata() if tool.metadata() else ToolMetadata()
    stdouts = [o.output_type for o in tool.outputs() if isinstance(o.output_type, Stdout) and o.output_type.stdoutname]
    stdout = stdouts[0].stdoutname if len(stdouts) > 0 else None

    tool_cwl = cwlgen.CommandLineTool(
        tool_id=tool.id(),
        base_command=tool.base_command(),
        label=tool.id(),
        doc=metadata.documentation,
        # cwl_version=Cwl.kCUR_VERSION,
        stdin=None,
        stderr=None,
        stdout=stdout
    )

    tool_cwl.requirements.extend([
        cwlgen.InlineJavascriptReq()
    ])

    if tool.requirements():
        tool_cwl.requirements.extend(tool.requirements())

    if with_docker:
        tool_cwl.requirements.append(cwlgen.DockerRequirement(
            docker_pull=tool.docker(),
            # docker_load=None,
            # docker_file=None,
            # docker_import=None,
            # docker_image_id=None,
            # docker_output_dir=None
        ))

    tool_cwl.inputs.extend(translate_tool_input(i) for i in tool.inputs())
    tool_cwl.outputs.extend(translate_tool_output(o) for o in tool.outputs())
    args = tool.arguments()
    if args:
        tool_cwl.arguments.extend(translate_tool_argument(a) for a in tool.arguments())

    return tool_cwl


def translate_input(inp):
    return cwlgen.InputParameter(
            param_id=inp.id(),
            label=inp.label,
            secondary_files=inp.data_type.secondary_files(),
            param_format=None,
            streamable=None,
            doc=inp.doc,
            input_binding=None,
            param_type=inp.data_type.cwl_type()
        )


def translate_output_node(node):
    return translate_output(node.output, next(iter(node.connection_map.values())).source())


def translate_output(outp, source):
    return cwlgen.WorkflowOutputParameter(
        param_id=outp.id(),
        output_source=source,
        label=outp.label,
        secondary_files=outp.data_type.secondary_files(),
        param_format=None,
        streamable=None,
        doc=outp.doc,
        param_type=outp.data_type.cwl_type(),
        output_binding=None,
        linkMerge=None
    )


def translate_tool_input(toolinput):
    default = toolinput.default if toolinput.default else toolinput.input_type.default()

    data_type = toolinput.input_type.cwl_type()
    input_binding = cwlgen.CommandLineBinding(
        # load_contents=toolinput.load_contents,
        position=toolinput.position,
        prefix=toolinput.prefix,
        separate=toolinput.separate_value_from_prefix,
        # item_separator=toolinput.item_separator,
        # value_from=toolinput.value_from,
        shell_quote=toolinput.shell_quote,
    )

    # Binding array inputs onto the console
    # https://www.commonwl.org/user_guide/09-array-inputs/
    if isinstance(toolinput.input_type, Array) and isinstance(data_type, cwlgen.CommandInputArraySchema):
        if toolinput.nest_input_binding_on_array:
            input_binding.prefix = None
            input_binding.separate = None
            nested_binding = cwlgen.CommandLineBinding(
                # load_contents=toolinput.load_contents,
                prefix=toolinput.prefix,
                separate=toolinput.separate_value_from_prefix,
                # item_separator=toolinput.item_separator,
                # value_from=toolinput.value_from,
                shell_quote=toolinput.shell_quote,
            )
            data_type.inputBinding = nested_binding
        else:
            input_binding.itemSeparator = ","

    return cwlgen.CommandInputParameter(
        param_id=toolinput.tag,
        label=toolinput.tag,
        secondary_files=toolinput.input_type.secondary_files(),
        # streamable=None,
        doc=toolinput.doc,
        input_binding=input_binding,
        default=default,
        param_type=data_type
    )


def translate_tool_argument(argument):
    if argument.value is None:
        val = None
    elif callable(getattr(argument.value, "cwl", None)):
        val = argument.value.cwl()
    else:
        val = str(argument.value)
    return cwlgen.CommandLineBinding(
        # load_contents=False,
        position=argument.position,
        prefix=argument.prefix,
        separate=argument.separate_value_from_prefix,
        # item_separator=None,
        value_from=val,
        shell_quote=argument.shell_quote,
    )

def translate_tool_output(output):
    return cwlgen.CommandOutputParameter(
        param_id=output.tag,
        label=output.tag,
        secondary_files=output.output_type.secondary_files(),
        # param_format=None,
        # streamable=None,
        doc=output.doc,
        output_binding=cwlgen.CommandOutputBinding(
            glob=output.glob,
            # load_contents=False,
            # output_eval=None
        ),
        param_type=output.output_type.cwl_type()
    )


def translate_step(step: StepNode, is_nested_tool=False):
    run_ref = ("{tool}.cwl" if is_nested_tool else "tools/{tool}.cwl").format(tool=step.step.tool().id())
    cwlstep = cwlgen.WorkflowStep(
        step_id=step.id(),
        run=run_ref,
        label=step.step.label,
        doc=step.step.doc,
        scatter=None,  # Filled by StepNode
        scatter_method=None  # Filled by StepNode
    )

    cwlstep.out = [cwlgen.WorkflowStepOutput(output_id=o.tag) for o in step.step.tool().outputs()]

    ins = step.inputs()
    scatterable = []

    for k in ins:
        inp = ins[k]
        if k not in step.connection_map:
            if inp.input_type.optional:
                continue
            else:
                raise Exception(f"Error when building connections for cwlstep '{step.id()}', "
                                f"could not find required connection: '{k}'")

        inp_t = step.inputs()[k].input_type
        edge = step.connection_map[k]
        default = edge.default if edge.default else inp_t.default()
        d = cwlgen.WorkflowStepInput(
            input_id=inp.tag,
            source=edge.source(),
            link_merge=None,  # this will need to change when edges have multiple source_map
            default=default,
            value_from=None
        )
        if edge.has_scatter():
            scatterable.append(k)

        cwlstep.inputs.append(d)

    if len(scatterable) > 0:
        if len(scatterable) > 1:
            Logger.info("Discovered more than one scatterable field on cwlstep '{step_id}', "
                        "deciding scatterMethod to be dot_product".format(step_id=step.id()))
            cwlstep.scatterMethod = "dot_product"
        cwlstep.scatter = scatterable

    return cwlstep


def generate_cwl_resource_override_schema():
    schema = cwlgen.CommandInputRecordSchema()
    schema.fields = [
        cwlgen.CommandInputRecordSchema.CommandInputRecordField("coresMin", ["long", "string", "null"]),
        cwlgen.CommandInputRecordSchema.CommandInputRecordField("coresMax", ["int", "string", "null"]),
        cwlgen.CommandInputRecordSchema.CommandInputRecordField("ramMin", ["long", "string", "null"]),
        cwlgen.CommandInputRecordSchema.CommandInputRecordField("ramMax", ["int", "string", "null"])
    ]
    return schema


def generate_cwl_resource_override_schema_for_steps(wf):
    from janis.workflow.workflow import Workflow
    schema = cwlgen.CommandInputRecordSchema()

    for step in wf._steps:
        tool = step.step.tool()
        if isinstance(tool, Workflow):
            key = tool.id() + "_resource_override"  # self.RESOURCE_OVERRIDE_KEY
            override_schema = generate_cwl_resource_override_schema_for_steps(tool)
        else:
            key = step.step.id()
            override_schema = generate_cwl_resource_override_schema()

        schema.fields.append(cwlgen.CommandInputRecordSchema.CommandInputRecordField(key, ["null", override_schema]))

    return schema


def generate_hint_selectors_for_hint_map(resource_key, hint_map):
    import json
    return f"""${{
var key = '{resource_key}';
if (inputs["{RESOURCE_OVERRIDE_KEY}"] && inputs["{RESOURCE_OVERRIDE_KEY}"][key])
    return inputs["{RESOURCE_OVERRIDE_KEY}"][key];
var hints = inputs.hints;
if (!hints) return null;
var hintMap = {json.dumps(hint_map)};
for (var hint in hintMap) {{
    var providedHintValue = hints[hint];
    if (!providedHintValue || !(providedHintValue in hintMap[hint])) continue;
    var hintValueToResourceMap = hintMap[hint][providedHintValue];
    if (hintValueToResourceMap[key]) return hintValueToResourceMap[key];
}}
return null;
}}"""  # .replace("    ", "").replace("\n", "")


def get_cwl_schema_for_recognised_hints():
    schema = cwlgen.CommandInputRecordSchema("hints")

    def prepare_hint(hint_class: Type[Hint]):
        name = hint_class.key() + "_schema"

        if issubclass(hint_class, HintEnum):
            # assume is an enum
            return cwlgen.CommandInputEnumSchema(label=hint_class.key(), name=name, symbols=hint_class.symbols())
        elif issubclass(hint_class, HintArray):
            return cwlgen.CommandInputArraySchema(items=hint_class.items(), label=hint_class.key())
        else:
            return "string?"

    schema.fields = [
        cwlgen.CommandInputRecordSchema.CommandInputRecordField(hint.key(), ["null", prepare_hint(hint)])
        for hint in HINTS]

    return schema



import warnings


class PipelineTranslatorException(Exception): pass


class DuplicateLabelIdentifier(Exception): pass


class InvalidNodeIdentifier(Exception): pass


class NodeNotFound(Exception): pass


class NotFoundException(Exception): pass


class InvalidInputsException(Exception): pass


class InvalidStepsException(Exception): pass


def deprecated(message):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn("{} is a deprecated function. {}".format(func.__name__, message),
                          category=DeprecationWarning,
                          stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)

        return deprecated_func

    return deprecated_decorator



from typing import List

from janis import ToolInput, File, ToolOutput, Array, String
from janis.tool.expressiontool import ExpressionTool
from janis.unix.tools.unixtool import UnixTool


class ScatterLines(ExpressionTool, UnixTool):

    @staticmethod
    def base_command():
        return None

    def friendly_name(self) -> str:
        return "Scatter across lines"

    @staticmethod
    def tool():
        return "parse-file"

    def expression(self):
        return """
${return { lines: inputs.file.contents.split("\n")
    .filter(function(q) { return q.length > 0; }) }}""".strip()

    def inputs(self) -> List[ToolInput]:
        return [
            ToolInput("file", File())
        ]

    def outputs(self):
        return [ToolOutput("lines", Array(String()))]



__version__ = "v0.2.7"

from janis.workflow.workflow import Workflow
from janis.workflow.step import Step
from janis.workflow.input import Input
from janis.workflow.output import Output
from janis.tool.commandtool import CommandTool, ToolInput, ToolOutput, ToolArgument
from janis.types.data_types import DataType
from janis.types.common_data_types import Boolean, String, Int, Float, Double, File, Directory, Array, Filename, Stdout
from janis.utils.logger import Logger, LogLevel
from janis.utils.metadata import Metadata, WorkflowMetadata, ToolMetadata
from janis.unix import *



# process for adding new hints!
from abc import ABC, abstractmethod
from typing import Union, List, Type


class Hint(ABC):
    # string inputs
    @staticmethod
    @abstractmethod
    def key():
        raise Exception("Must implement key() method")


class HintEnum(Hint, ABC):
    @staticmethod
    @abstractmethod
    def symbols():
        raise Exception("Must implement symbols() method")


class HintArray(Hint, ABC):
    @staticmethod
    @abstractmethod
    def items():
        raise Exception("Must implement items() method")


class CaptureType(HintEnum):

    KEY = "captureType"
    TARGETED = "targeted"
    EXOME = "exome"
    THIRTYX = "30x"
    NINETYX = "90x"
    THREEHUNDREDX = "300x"
    ALL = [TARGETED, EXOME, THIRTYX, NINETYX, THREEHUNDREDX]

    @staticmethod
    def key(): return CaptureType.KEY
    @staticmethod
    def symbols(): return CaptureType.ALL

HINTS: List[Type[Hint]] = [CaptureType]

def get_cwl_schema_for_recognised_hints():
    import cwlgen.cwlgen as cwl

    schema = cwl.CommandInputRecordSchema("hints")

    def prepare_hint(hint_class: Type[Hint]):
        name = hint_class.key() + "_schema"

        if issubclass(hint_class, HintEnum):
            # assume is an enum
            return cwl.CommandInputEnumSchema(label=hint_class.key(), name=name, symbols=hint_class.symbols())
        elif issubclass(hint_class, HintArray):
            return cwl.CommandInputArraySchema(items=hint_class.items(), label=hint_class.key())
        else:
            return "string?"

    schema.fields = [
        cwl.CommandInputRecordSchema.CommandInputRecordField(hint.key(), ["null", prepare_hint(hint)])
        for hint in HINTS]

    return schema



from weakref import WeakKeyDictionary
from typing import Any
from janis.utils.logger import Logger


class BaseDescriptor(object):
    def __init__(self, default: Any = None):
        self.default: Any = default
        self.data: WeakKeyDictionary = WeakKeyDictionary()

    def __get__(self, instance, owner):
        # Logger.log(f"Getting value from {instance} and {owner}")
        return self.data.get(instance, self.default)

    def __set__(self, instance, value):
        # Logger.log(f"Setting value {value} to {instance}")
        self.data[instance] = value


class GetOnlyDescriptor(object):
    def __init__(self, default: Any = None):
        self.default: Any = default
        self.data: WeakKeyDictionary = WeakKeyDictionary()

    def __get__(self, instance, owner):
        Logger.log(f"Getting readonly value from {instance} and {owner}")
        return self.data.get(instance, self.default)

    def __set__(self, instance, value):
        if not self.data[instance]:
            return Logger.log("Attempting to write to read-only property")

        Logger.log(f"Setting readonlu value {value} to {instance}")
        self.data[instance] = value



from abc import ABC, abstractmethod
import re
from typing import List, Dict, Optional, Any

from janis.tool.tool import Tool, ToolArgument, ToolInput, ToolOutput, ToolTypes

import cwlgen as cwl
from janis.types.common_data_types import Stdout, Array
from janis.utils import convert_expression_to_wdl
from janis.utils.logger import Logger
from janis.utils.metadata import Metadata
from janis.utils.validators import Validators


class CommandTool(Tool, ABC):
    """
    Notes:
        - If you're thinking about secondary files, DON'T!! Consider creating a new DataType.
        - This class is similar to how RABIX COMPOSER creates the tools
        - You can subclass and override whichever fields you'd like, including the INPUTS / OUTPUTS!
        - Take note which options you can provide to the ToolInput and ToolOutput.
    """

    def id(self):
        return self.tool().lower()

    @staticmethod
    @abstractmethod
    def tool():
        raise Exception(f"subclass MUST implement 'tool' method")

    @classmethod
    def full_name(cls):
        if cls.version() is not None:
            return f"{cls.tool()}/{cls.version()}"
        return cls.tool()

    @staticmethod
    def docker():
        return "ubuntu:latest"

    @staticmethod
    @abstractmethod
    def base_command():
        raise Exception("Subclass MUST implement the 'base_command' method with str or [str] return types")

    def memory(self) -> Optional[int]:
        return None

    def cpus(self) -> Optional[int]:
        return None

    def arguments(self) -> Optional[List[ToolArgument]]:
        return None

    @classmethod
    def type(cls):
        return ToolTypes.CommandTool

    @staticmethod
    def environment_variables() -> Optional[Dict[str, str]]:
        return None

    @staticmethod
    def requirements() -> Optional[List[cwl.Requirement]]:
        return None

    @staticmethod
    def hint_map() -> Optional[Dict[str, Any]]:
        return None

    def wdl(self, with_docker=True):
        import wdlgen as wdl

        # Todo: Move this to python-wdlgen
        if not Validators.validate_identifier(self.id()):
            raise Exception(f"The identifier '{self.id()}' for class '{self.__class__.__name__}' was not validated by "
                            f"'{Validators.identifier_regex}' (must start with letters, and then only contain letters, "
                            f"numbers or an underscore)")

        ins, outs = [], []
        for i in self.inputs():
            wd = i.input_type.wdl()
            if isinstance(wd, list):
                ins.extend(wdl.Input(w, i.id()) for w in wd)
            else:
                ins.append(wdl.Input(wd, i.id()))

        for o in self.outputs():

            if isinstance(o.output_type, Stdout):
                expression = "stdout()"
            else:

                glob = convert_expression_to_wdl(o.glob)
                if glob is not None and "*" in glob:
                    glob = f'glob({glob})'
                    if not isinstance(o.output_type, Array):
                        Logger.warn(f"The command tool '{self.id()}.{o.tag}' used a star-bind (*) glob to find the output, "
                                    f"but the return type was not an array. For WDL, the first element will be used, "
                                    f"ie: '{glob}[0]'")
                        glob = glob + "[0]"
                expression = glob

            outs.append(wdl.Output(o.output_type.wdl(), o.id(), expression))

        command_ins = [

            wdl.Task.Command.CommandInput(
                name=i.id(),
                optional=i.input_type.optional,
                prefix=i.prefix,
                position=i.position,
                separate_value_from_prefix=i.separate_value_from_prefix,
                default=i.default if i.default else i.input_type.default()
            ) for i in self.inputs()] if self.inputs() else None

        command_args = None
        if self.arguments():
            command_args = []
            for a in self.arguments():
                if a.value is None:
                    val = None
                elif callable(getattr(a.value, "wdl", None)):
                    val = a.value.wdl()
                else:
                    val = a.value
                command_args.append(wdl.Task.Command.CommandArgument(a.prefix, val, a.position))

        command = wdl.Task.Command(self.base_command(), command_ins, command_args)

        r = wdl.Task.Runtime()
        if with_docker:
            r.add_docker(self.docker())

        return wdl.Task(self.id(), ins, outs, command, r)


    def help(self):
        import inspect
        path = inspect.getfile(self.__class__)

        ins = sorted(self.inputs(), key=lambda i: i.position if i.position is not None else 0)
        args = ""
        if self.arguments():
            args = " " + " ".join(f"{(a.prefix if a.prefix is not None else '') + ' ' if (a.prefix is not None and a.separate_value_from_prefix) else ''}{a.value}" for a in self.arguments())

        prefixes = " -" + "".join(i.prefix.replace("-", "").replace(" ", "") for i in ins if i.prefix is not None)

        metadata = self.metadata() if self.metadata() else Metadata()
        docker = self.docker()

        base = (self.base_command() if isinstance(self.base_command(), str) else " ".join(self.base_command())) \
            if self.base_command() else ''
        command = base + args + prefixes

        def input_format(t: ToolInput):
            prefix_with_space = ""
            if t.prefix is not None:
                prefix_with_space = (t.prefix + ": ") if (t.separate_value_from_prefix is not False) else t.prefix
            return f"\t\t{t.tag} ({prefix_with_space}{t.input_type.id()}{('=' + str(t.default)) if t.default is not None else ''})" \
                f": {'' if t.doc is None else t.doc}"

        output_format = lambda t: f"\t\t{t.tag} ({t.output_type.id()}): {'' if t.doc is None else t.doc}"

        requiredInputs = "\n".join(input_format(x) for x in ins if not x.optional)
        optionalInputs = "\n".join(input_format(x) for x in ins if x.optional)
        outputs = "\n".join(output_format(o) for o in self.outputs())

        return f"""
    Pipeline tool: {path} ({self.id()})
NAME
    {self.id()}

SYNOPSIS
    {command}

DOCKER
    {docker}

DOCUMENTATION URL
    {metadata.documentationUrl if metadata.documentationUrl else "No url provided"}

DESCRIPTION
    {metadata.documentation if metadata.documentation else "No documentation provided"}

INPUTS:
    REQUIRED:
{requiredInputs}

    OPTIONAL:
{optionalInputs}

OUTPUTS:
{outputs}
"""



from janis.tool.commandtool import ToolInput
from janis.types.common_data_types import Array, File
from janis.unix.tools.unixtool import UnixTool


class Cat(UnixTool):

    @staticmethod
    def tool():
        return "cat"

    def friendly_name(self):
        return "Concatenate"

    @staticmethod
    def base_command():
        return "cat"

    def inputs(self):
        return [
            ToolInput("files", Array(File()))
        ]

    def outputs(self):
        return []



from typing import Dict, Optional, List

from janis.utils.logger import Logger, LogLevel
# import cwlgen.cwlgen as cwl
from janis.types.data_types import DataType, NativeTypes
from janis.workflow.step import ToolOutput
from janis.graph.node import Node, NodeTypes


class Input:
    def __init__(self, identifier: str, data_type: DataType, value_from_input=None,
                 label: str=None, doc: str=None):
        Logger.log(f"Creating input '{identifier}' with value: '{value_from_input}'")
        self._identifier: str = identifier

        self.label = label
        self.doc = doc

        self.value = value_from_input

        self.data_type: DataType = data_type

    def id(self):
        return self._identifier

    def cwl_input(self):
        return self.data_type.cwl_input(self.value)

    def wdl_input(self):
        return self.value


class InputNode(Node):

    def __init__(self, inp: Input):
        Node.__init__(self, NodeTypes.INPUT, inp.id())
        self.input: Input = inp

    def outputs(self) -> Dict[str, ToolOutput]:
        # Program will just grab first value anyway
        return {"": ToolOutput(self.input._identifier, self.input.data_type)}

    def inputs(self):
        return None



from typing import List

import janis.hints as hints
from janis import ToolOutput, ToolInput, String, Stdout, File
from janis.unix.tools.unixtool import UnixTool


class DebugEchoInputs(UnixTool):
    @staticmethod
    def tool():
        return "debuginputs"

    @staticmethod
    def base_command():
        return "echo"

    def inputs(self) -> List[ToolInput]:
        return [
            # add whatever you want here
            ToolInput("stringInput", String(optional=True)),
            ToolInput("fileInput", File(optional=True))
        ]

    def outputs(self) -> List[ToolOutput]:
        return [
            ToolOutput("output", Stdout())
        ]

    def friendly_name(self) -> str:
        return "Debug inputs (log)"

    @staticmethod
    def hint_map():
        return {
            hints.CaptureType.KEY: {
                hints.CaptureType.TARGETED: {
                    "ramMin": 2000,
                    "coresMin": 3
                }
            }
        }



import json
from _sha1 import sha1
from datetime import date


class Metadata(object):

    DATE_FORMAT = "%Y-%m-%d"

    def __init__(self, creator=None, maintainer=None, maintainer_email=None, date_created=None, date_updated=None,
                 institution=None, doi=None, citation=None, keywords=None, documentation_url=None, documentation=None,
                 version=None):
        """

        :param creator:
        :param maintainer:
        :param maintainer_email:
        :param date_created:
        :param date_updated:
        :param institution:
        :param doi:
        :param citation:
        :type citation: str | list[str] | None
        :param keywords:
        :param documentation_url:
        :param documentation:
        :param version:
        """
        self.creator = creator
        self.maintainer = maintainer
        self.maintainerEmail = maintainer_email
        self.dateCreated = date_created.strftime(self.DATE_FORMAT) if isinstance(date_created, date) else date_created
        self.dateUpdated = date_created.strftime(self.DATE_FORMAT) if isinstance(date_updated, date) else date_updated
        self.institution = institution
        self.doi = doi
        self.citation = citation
        self.keywords = keywords
        self.documentation = documentation
        self.documentationUrl = documentation_url
        self.version = version

    def update(self, **kwargs):
        for k in kwargs:
            self.__setattr__(k, kwargs[k])
        return self

    def get_dict(self, object_to_checksum):

        checksum = sha1(json.dumps(object_to_checksum, sort_keys=True).encode('utf-8')).hexdigest()
        # https://stackoverflow.com/q/5884066

        d = {k: v for k, v in vars(self).items() if v is not None}
        d["checksum"] = checksum
        d["dateGenerated"] = date.today().strftime(self.DATE_FORMAT)

        return d


class WorkflowMetadata(Metadata):
    def get_dict(self, object_to_checksum, ninputs=None, nsteps=None, noutputs=None):
        d = super(WorkflowMetadata, self).get_dict(object_to_checksum)

        if ninputs:
            d["numberOfInputs"] = ninputs
        if nsteps:
            d["numberOfSteps"] = nsteps
        if noutputs:
            d["numberOfOutputs"] = noutputs

        return d


class ToolMetadata(Metadata):
    pass




from typing import Dict, List, Tuple, Optional, Any, Union

import networkx as nx
import wdlgen as wdl

from janis.translations import SupportedTranslation
from janis.utils.janisconstants import RESOURCE_OVERRIDE_KEY, HINTS_KEY
from janis.graph.node import Node, NodeTypes, layout_nodes2
from janis.graph.stepinput import StepInput
from janis.tool.tool import Tool, ToolInput, ToolOutput, ToolTypes, ToolType
from janis.types.common_data_types import Array
from janis.types.data_types import DataType
from janis.utils import first_value
from janis.utils.errors import DuplicateLabelIdentifier, InvalidNodeIdentifier, NodeNotFound, InvalidStepsException, \
    InvalidInputsException
from janis.utils.logger import Logger, LogLevel
from janis.utils.metadata import WorkflowMetadata
from janis.utils.validators import Validators
from janis.workflow.input import Input, InputNode
from janis.workflow.output import Output, OutputNode
from janis.workflow.step import Step, StepNode
import janis.translations as translations


class Workflow(Tool):
    """
    Documentation here
    """

    # identifier = BaseDescriptor()
    # label = BaseDescriptor()
    # doc = BaseDescriptor()

    def __init__(self, identifier: str, friendly_name: str = None, doc: Optional[str] = None):
        """
        Initialise the workflow
        :param identifier: uniquely identifies the workflow
        :param friendly_name: a label that the engine may use to represent the workflow (should be unique)
        :param doc: Documentation of the workflow
        """
        Logger.log(f"Creating workflow with identifier: '{identifier}'")

        if not Validators.validate_identifier(identifier):
            raise Exception(f"The identifier '{identifier}' was not validated by '{Validators.identifier_regex}' "
                            f"(must start with letters, and then only contain letters, numbers and an underscore)")

        self.identifier = identifier
        self.name = friendly_name
        self._metadata = WorkflowMetadata(documentation=doc)

        self._nodes: Dict[str, Node] = {}  # Look up a node by its identifier

        self._inputs: List[InputNode] = []  # InputNodes
        self._steps: List[StepNode] = []  # StepNodes
        self._outputs: List[OutputNode] = []  # OutputNodes
        self._outputs: List[OutputNode] = []  # OutputNodes

        self.graph: nx.MultiDiGraph = nx.MultiDiGraph()  # Realistically this isn't really used except for an image

        # Flags for different requirements that a workflow might need
        self.has_scatter = False
        self.has_subworkflow = False
        self.has_multiple_inputs = False

    def id(self) -> str:
        """
        Returns the identifier of the workflow
        """
        return self.identifier

    def friendly_name(self):
        return self.name

    def metadata(self):
        return self._metadata

    def doc(self):
        return self._metadata.documentation

    @classmethod
    def type(cls) -> ToolType:
        """
        Returns the ToolType (Workflow | CommandTool | ExpressionTool)
        """
        return ToolTypes.Workflow

    def inputs(self) -> List[ToolInput]:
        """
        List of ToolInputs of the workflow, we can toss out most of the metadata
        about positioning, prefixes, etc that the ToolInput class uses
        """
        return [ToolInput(i.id(), i.input.data_type) for i in self._inputs]

    def outputs(self) -> List[ToolOutput]:
        """
        Similar to inputs, return a list of ToolOutputs of the workflow
        """
        return [ToolOutput(o.id(), o.output.data_type, o.output.doc)
                for o in self._outputs if o.output.data_type is not None]

    # GRAPH CONSTRUCTION

    def draw_graph(self):
        """
        Use matplotlib to draw the a diagram of the workflow (currently neglects subworkflows).
        This will pause the execution of the program until the graph is closed.
        """
        import matplotlib.pyplot as plt

        default_color = 'black'

        G = self.graph
        edges_attributes = [G.edges[e] for e in G.edges]
        edge_colors = [x["color"] if 'color' in x else default_color for x in edges_attributes]
        node_colors = [NodeTypes.to_col(x.node_type) for x in G.nodes]

        pos = layout_nodes2(list(G.nodes))  # Manual layout engine of the graph, need to make this better

        for n in pos:
            G.node[n]['pos'] = pos[n]

        nx.draw(G, pos=pos, edge_color=edge_colors, node_color=node_colors, with_labels=True)
        plt.show()

    def _add_items(self, *args) -> List[Any]:
        """
        Supports adding (Input | Step | Output | Tuple = construct edge)
        """
        mixed: List[Any] = args

        if isinstance(args[0], list):
            if len(args) > 1:
                raise Exception("Invalid type of arguments")
            mixed = args[0]

        return [self._add_item(i) for i in mixed]

    def _add_item(self, item: Union[Input, Step, Output, Tuple]):
        """
        Generic add whatever into the workflow: (Input | Step | Output | Tuple = construct edge)
        :return:
        """
        if isinstance(item, Input):
            return self._add_inputs([item])[0]
        elif isinstance(item, Step):
            return self._add_steps([item])[0]
        elif isinstance(item, Output):
            return self._add_outputs([item])[0]
        elif isinstance(item, tuple):
            return self.add_edge(item[0], item[1])
        else:
            raise Exception(f"Unexpected type '{type(item)}' passed to 'add_nodes'")

    def _add_inputs(self, inputs: List[Input]) -> List[InputNode]:
        """
        Add a list of inputs into the graph, does nothing except add it into the graph
        """
        ins = []
        for inp in inputs:
            Logger.log(f"Adding input '{inp.id()}' to '{self.identifier}'")
            node: InputNode = InputNode(inp)
            ins.append(node)
            self._add_node(node)
            self._inputs.append(node)
        return ins

    def _add_outputs(self, outputs: List[Output]) -> List[OutputNode]:
        """
        Adds a list of output to the graph, and returns the created OutputNodes
        """
        outs = []
        for outp in outputs:
            Logger.log(f"Adding output '{outp.id()}' to '{self.identifier}'")
            node: OutputNode = OutputNode(outp)
            outs.append(node)
            self._add_node(node)
            self._outputs.append(node)
        return outs

    def _add_steps(self, steps: List[Step]) -> List[StepNode]:
        """
        Adds a list of steps to the graph and returns the created StepNodes
        """
        ss = []
        for step in steps:
            Logger.log(f"Adding step '{step.id()}' to '{self.identifier}'")
            node: StepNode = StepNode(step)

            self.has_subworkflow = self.has_subworkflow or step.tool().type() == ToolTypes.Workflow
            ss.append(node)
            self._add_node(node)
            self._steps.append(node)
        return ss

    def _add_node(self, node: Node):
        """
        Add a node to the graph, does the correct error checking
        """
        if node.id() in self._nodes:
            existing = self._nodes[node.id()]
            message = f"Attempted to add an input with the identifier '{node.id()}' but there was already " \
                f"an {existing.node_type} node with representation: {str(existing)}"
            Logger.log(message)
            raise DuplicateLabelIdentifier(message)
        self._nodes[node.id()] = node
        self.graph.add_node(node)
        return node

    def add_items(self, items: List[Union[Input, Step, Output, Tuple]]):
        return [self._add_item(n) for n in items]

    @staticmethod
    def get_labels_and_node_by_inference(s) -> Tuple[str, Optional[Node]]:
        """
        Try to get the identifier and node of the
        :return: Tuple[str, Optional[Node]]
        """
        try:
            if type(s) == str:
                # just a string was passed,
                return s, None
            elif type(s) == tuple:
                # should have format (tag, Optional[Node]), check 105:step.py
                return s
            else:
                # s is hopefully a node (either InputNode | OutputNode)
                return s.id(), s
        except Exception as e:
            Logger.log_ex(e)
            return str(s), None

    def connect_inputs(self, step: Any, inputs: List[Any]) -> List[StepInput]:
        """
        Connect all the inputs to the step (infer the connections, will error if it can't be determined)
        :param step: Step to connect to
        :param inputs: Inputs to connect from
        :return: List[Edge]
        """
        return [self.add_piped_edge(step, i) for i in inputs]

    def add_edges(self, edges: List[Tuple[Any, Any]]):
        return [self.add_edge(e[0], e[1]) for e in edges]

    def add_pipe(self, *args) -> List[StepInput]:
        """
        Connect args[0] -> args[1] and args[1] to args[2] and ... and args[n-1] to args[n]
        :return:
        """
        if len(args) < 2:
            raise Exception("Must pass at least two properties to 'add_pipe'")

        # add_piped_edge removes the component on the start node (ie: n0 -> n1.tag, n1 -> n2.tag2, etc)
        return [self.add_piped_edge(args[i], args[i + 1]) for i in range(len(args) - 1)]

    def add_piped_edge(self, start, finish) -> StepInput:
        """
        add_piped_edge removes the component on the start node (ie: n0 -> n1.tag, n1 -> n2.tag2, etc)
        :param start: node to connect from
        :param finish: node to connect to
        :return: Edge between start and finish
        """
        s, sn = self.get_labels_and_node_by_inference(start)
        f, fn = self.get_labels_and_node_by_inference(finish)
        return self.add_edge((s.split("/")[0], sn), (f, fn))

    def add_default_value(self, tagged_node, default_value):
        """
        Adds a default value to the edge, will create an edge!
        :param tagged_node: will be run through get_labels_and_node_by_inference: ( string | (string, node))
        :param default_value: value to be the default, should match the data_type
        """
        if not tagged_node:
            raise Exception("You must pass a non-optional node to 'add_default_value'")
        n, component = self.get_labels_and_node_by_inference(tagged_node)

        if not n:
            raise Exception(f"Couldn't resolve the node identifier from '{str(tagged_node)}'")
        components = n.split("/")
        node_id = components[0]
        node = self._nodes.get(node_id)
        if not node:
            raise Exception(f"Couldn't find a node in the graph with identifier '{node_id}', you must add the node "
                            f"to the graph to add a default value")
        if node.node_type != NodeTypes.TASK:
            raise Exception(f"You can only add a default value to a task (step) (not to '{node.id()}'")
        if not component:
            has_one_value = ", even if there's only one value" if len(node.inputs()) == 1 else ""
            raise Exception("You must fully qualify a node and it's components to add a default value" + has_one_value)

        tag = components[-1]

        # # Validate the type of data?
        # inp = node.inputs().get(tag)
        # inp.input_type == type(default)

        if tag in node.connection_map:
            e = node.connection_map[tag]
        else:
            e = StepInput(node, tag)
            node.connection_map[components[-1]] = e
        e.set_default(default_value)

    @staticmethod
    def validate_tag_parts(parts: List[str], tag: str, node_type: str):
        """
        Validates the parts in a tag (between 0 and 2)
        :param parts: list
        :param tag:
        :param node_type: start | finish for the error string
        :return:
        """
        if len(parts) == 0 or len(parts) > 2:
            message = f"The {node_type} tag '{tag}' contained an invalid number of parts ({len(parts)}), " \
                f"it should follow the format: 'label/tag' (1 or 2 parts)"
            Logger.log(message, LogLevel.CRITICAL)
            raise InvalidNodeIdentifier(message)

    def try_get_or_add_node(self, identifier: str, component: Optional[Union[Input, Step, Output]],
                            node_type: str) -> Node:
        """
        Given the label (and the component [Input \ Step | Output], get the node if it's in the graph
        (and check that it's referencing the correct component), else add it to the graph.
        :param identifier: identifier
        :type identifier: str
        :param component: the step | input | output that the node contains
        :type component: Input | Step | Output
        :param node_type: Used to better classify errors ["input" | "step" | "output"]
        :type node_type: str
        :return: Node | Exception
        :returns: Node
        """
        node = self._nodes.get(identifier)

        if node is not None:
            # Try to do some proper matching
            node_comp = None
            if isinstance(node, InputNode):
                node_comp = node.input
            elif isinstance(node, StepNode):
                node_comp = node.step
            elif isinstance(node, OutputNode):
                node_comp = node.output

            if component is not None and (
                    identifier != component.id() or
                    (isinstance(component, Input) and not node.node_type == NodeTypes.INPUT) or
                    (isinstance(component, Step) and not node.node_type == NodeTypes.TASK) or
                    (isinstance(component, Output) and not node.node_type == NodeTypes.OUTPUT) or
                    (node_comp != component)):
                raise Exception(f"There already exists a node (and component) with id '{node.id()}'. The added "
                                f"component ('{repr(component)}') clashes with '{repr(node_comp)}').")
            return node

        Logger.log(f"Could't find a(n) {node_type} node with the identifier '{identifier}' in the workflow.")
        if component is not None:
            Logger.log(f"Adding '{component.id()}' to the workflow")
            return self._add_item(component)
        else:
            message = f"There was no node or component referenced by '{identifier}' in the graph. When creating edges " \
                "by identifiers, you must add the component to the graph first"
            Logger.log(message, LogLevel.CRITICAL)
            raise NodeNotFound(message)

    def add_edge(self, start, finish, default_value=None):
        """
        Add the start node as a source to the finish node, will create an edge if none exists
        :param start: an input or step node pair (via get_labels_and_node_by_inference)
        :param finish: a step or output node pair (via get_labels_and_node_by_inference)
        :param default_value: optionally provide a default value, if the start resolves to null
        :return: edge
        """

        s, s_component = self.get_labels_and_node_by_inference(start)
        f, f_component = self.get_labels_and_node_by_inference(finish)
        Logger.log(f"Adding edge between {s} and {f}")

        # set nodes
        s_parts = s.split("/")
        f_parts = f.split("/")
        stag, ftag = None, None

        # ORDER:
        #   1. Get the input type to link
        #   2. Try and find this type in the cur_step body
        #       (a) - If it's optional, skip
        #       (b) - else: throw error
        #   3. Determine where the input is coming from (input / other cur_step)
        #   4. Try to exactly connect the input to the output/tag
        #       (a) If there's only one output, log warning if it doesn't exactly match
        #       (b) If there are multiple outputs, raise Exception if it doesn't exactly match
        #   5. Check that the types match
        #   6. Add edge
        self.validate_tag_parts(s_parts, s, "start")
        self.validate_tag_parts(f_parts, f, "finish")

        s_label = s_parts[0]
        f_label = f_parts[0]

        s_node = self.try_get_or_add_node(s_label, s_component, "start")
        f_node = self.try_get_or_add_node(f_label, f_component, "finish")

        s_node.set_depth(f_node.depth - 1)
        f_node.set_depth(s_node.depth + 1)

        if s_node.node_type == NodeTypes.INPUT:
            # Can guarantee the data_type
            s_parts, s_type = self.get_tag_and_type_from_start_edge_node(s_node, s_parts, referenced_by=f_parts)
            f_parts, f_type = self.get_tag_and_type_from_final_edge_node(f_node, f_parts, guess_type=s_type)
            stag, ftag = None, f_parts[-1]

        elif f_node.node_type == NodeTypes.OUTPUT:
            # We'll need to determine the data_type and assign it to the output
            s_parts, s_type = self.get_tag_and_type_from_start_edge_node(s_node, s_parts, referenced_by=f_parts)
            if s_type is None:
                keys = ", ".join(s_node.outputs().keys())
                raise Exception(f"The step '{s_node.id()}' must be fully qualified to connect to output node"
                                f" (expected one of: {keys})")

            step_has_scatter = any(e.has_scatter() for e in s_node.connection_map.values())

            if step_has_scatter:
                f_type = Array(s_type.received_type())
            else:
                f_type = s_type.received_type()

            f_node: OutputNode = f_node
            f_parts = [f_label]
            f_node.output.data_type = f_type
            stag, ftag = s_parts[-1], None
            Logger.log(f"Connecting '{s_node.id()}' to output '{f_node.id()}'")
        else:

            # We can't guarantee or enforce anything, so we'll see if we are fully qualified, and if not we can try
            # and guess by comparing types between the nodes, it doesn't matter which way we'll start though
            s_parts, s_type = self.get_tag_and_type_from_start_edge_node(s_node, s_parts, referenced_by=f_parts)
            f_parts, f_type = self.get_tag_and_type_from_final_edge_node(f_node, f_parts, guess_type=s_type)

            if s_type is None and f_type is not None:
                s_parts, s_type = self.get_tag_and_type_from_start_edge_node(
                    s_node,
                    s_parts,
                    referenced_by=f_parts,
                    guess_type=f_type
                )

            if s_type is None and f_type is None:
                # Try to guess both
                Logger.log(f"Trying to guess connection based on mutual types between "
                           f"'{s_node.id()}' and '{f_node.id()}'", LogLevel.WARNING)
                c = self.guess_connection_between_nodes(s_node, f_node)
                if c is not None:
                    (s_parts, s_type), (f_parts, f_type) = c

            stag, ftag = s_parts[-1], f_parts[-1]

        # We got one but not the other
        if s_type is not None and f_type is None:
            possible_outp_tags = ", ".join(f"{o.tag} ({o.input_type.name()})" for o in f_node.inputs().values())
            raise Exception(f"Couldn't definitively establish a connection from the start node '{s_node.id()}' "
                            f"with type {s_type.name()} to the finish node with possibilities: {possible_outp_tags}")
        if s_type is None and f_type is not None:
            possible_inp_tags = ", ".join(f"{i.tag} ({i.output_type.name()})" for i in s_node.outputs().values())
            raise Exception(f"Couldn't definitively establish a connection to the finish node '{f_node.id()}' "
                            f"with type {f_type.name()} from the start node with possibilities: {possible_inp_tags}")

        # We couldn't get either
        if s_type is None or f_type is None:
            ss = '/'.join(s_parts)
            ff = '/'.join(f_parts)
            raise Exception(f"Couldn't connect '{s_node.id()}' to '{f_node.id()}', failed to get types from "
                            f"tags (s: {ss} | f: {ff}) and couldn't uniquely isolate mutual types")

        # NOW: Let's build the connection (edge)

        if ftag in f_node.connection_map:
            step_inputs = f_node.connection_map[ftag]
        else:
            step_inputs = StepInput(f_node, ftag)
            f_node.connection_map[ftag] = step_inputs

        e = step_inputs.add_source(s_node, stag)
        if default_value:
            step_inputs.set_default(default_value)

        self.has_scatter = self.has_scatter or e.scatter
        self.has_multiple_inputs = self.has_multiple_inputs or step_inputs.multiple_inputs

        col = 'black' if e.compatible_types else 'r'
        self.graph.add_edge(s_node, f_node, type_match=e.compatible_types, color=col)
        return e

    def get_tag_and_type_from_start_edge_node(self, node: Node, input_parts: List[str], referenced_by: List[str],
                                              guess_type: Optional[DataType] = None) \
            -> Tuple[List[str], Optional[DataType]]:

        if guess_type is not None:
            guess_type = guess_type.received_type()

        if node.node_type == NodeTypes.OUTPUT:
            raise Exception(f"Can't join output node '{node.id()}' to '{referenced_by}'")

        lbl = input_parts[0]
        if node.node_type == NodeTypes.INPUT:
            if len(input_parts) != 1:
                s = "/".join(input_parts)
                message = f"The input tag '{s}' over-represents the INPUT node, this was corrected ({s} → {lbl})"
                Logger.log(message, LogLevel.WARNING)
                input_parts = [lbl]
            return input_parts, next(iter(node.outputs().values())).output_type

        # We are a step node now
        if not isinstance(node, StepNode):
            raise Exception("An internal error has occurred in 'get_tag_and_type_from_start_edge_node'"
                            " when converting Node to StepNode.")
        snode: StepNode = node
        outs: Dict[str, ToolOutput] = node.outputs()

        types = [(x, outs[x].output_type) for x in outs]

        if len(types) == 0:
            raise InvalidStepsException(f"The step '{referenced_by}' referenced the step '{snode.id()}' with tool "
                                        f"'{snode.step.tool().id()}' that has no outputs")
        elif len(types) == 1:
            tag = types[0][0]
            if len(input_parts) != 2:
                s = "/".join(input_parts)
                under_over = "under" if len(input_parts) < 2 else "over"
                Logger.warn(f"The node '{'/'.join(referenced_by)}' {under_over}-referenced an output of the tool "
                            f"'{snode.step.tool().id()}' (step: {node.id()}, this was automatically corrected "
                            f"({s} → {lbl}/{tag})")
            elif input_parts[-1] != tag:
                Logger.log(f"The node '{node.id()}' did not correctly reference an output of the tool "
                           f"'{snode.step.tool().id()}', this was automatically corrected "
                           f"({'/'.join(input_parts)} → {lbl}/{tag})", LogLevel.WARNING)

            input_parts = [lbl, tag]
            return input_parts, types[0][1]

        else:
            # if the edge tag doesn't match, we can give up
            if len(input_parts) < 2:
                if guess_type is not None:
                    # try to guess it from the outputs
                    compatible_types = [outs[x] for x in outs if (not outs[x].output_type.optional) and guess_type.can_receive_from(outs[x].output_type)]
                    if len(compatible_types) == 1:
                        out = compatible_types[0]
                        Logger.info(
                            f"Guessed the compatible match for '{node.id()}' with source type '{out.output_type.id()}'"
                            f" → '{guess_type.id()}' ('{node.id()}/{out.tag}' → '{node.id()}')")
                        return [lbl, out.tag], out.output_type
                    else:
                        # Should this step also take into consideration the _optional_ nature of the node,
                        # ie: should it favour required nodes if that's an option
                        ultra_compatible = [outs[x] for x in outs if (not outs[x].output_type.optional) and type(outs[x].output_type) == type(guess_type)]

                        if len(ultra_compatible) == 1:
                            ultra = ultra_compatible[0]
                            Logger.warn(f"There were {len(compatible_types)} matched types for the node '{node.id()}', "
                                        f"the program has guessed an exact compatible match of "
                                        f"type '{ultra.output_type.id()}' to tag '{ultra.tag}'")
                            return [lbl, ultra.tag], ultra.output_type
                        else:
                            s = "/".join(input_parts)
                            compat_str = ", ".join(f"{x.tag}: {x.output_type.id()}" for x in compatible_types)
                            raise Exception(
                                f"The node '{node.id()}' did not specify an input tag, and used '{guess_type.id()}'"
                                f" from the start node to guess the input by type, matching {len(compatible_types)}"
                                f" compatible ({compat_str}) and {len(ultra_compatible)} exact types."
                                f" You will need to provide more information to proceed.")
                else:
                    possible_tags = ", ".join(f"'{x}'" for x in outs)
                    s = "/".join(input_parts)
                    Logger.critical(
                        f"The tag '{s}' could not uniquely identify an input of '{snode.id()}', requires the "
                        f"one of the following tags: {possible_tags}")
                    return input_parts, None

            tag = input_parts[1]
            t = snode.step.tool().outputs_map().get(tag)
            if t:
                if len(input_parts) != 2:
                    Logger.log(f"The node '{node.id()}' did not correctly reference an output of the tool "
                               f"'{snode.step.tool().id()}', this was automatically corrected "
                               f"({'/'.join(input_parts)} → {lbl}/{tag})", LogLevel.WARNING)
                input_parts = [lbl, tag]
                return input_parts, t.output_type

            possible_tags = ", ".join(f"'{x}'" for x in outs)
            raise Exception(f"Could not identify an output called '{tag}' on the node '{node.id()}' "
                            f", possible tags: {possible_tags}")

        return input_parts, None

    @staticmethod
    def get_tag_and_type_from_final_edge_node(node: Node, input_parts: List[str],
                                              guess_type: Optional[DataType]) -> Tuple[List[str], Optional[DataType]]:
        if guess_type is not None:
            guess_type = guess_type.received_type()

        if node.node_type == NodeTypes.INPUT:
            raise Exception(f"Can't connect TO input node '{node.id()}'")
        if node.node_type == NodeTypes.OUTPUT:
            raise Exception("An internal error has occurred: output nodes should be filtered from the "
                            "function 'get_tag_and_type_from_final_edge_node'")

        if not isinstance(node, StepNode):
            raise Exception("An internal error has occurred in 'get_tag_and_type_from_start_edge_node'"
                            " when converting Node to StepNode.")

        snode: StepNode = node
        ins = snode.inputs()
        types = [(x, ins[x].input_type) for x in ins]
        lbl = input_parts[0]

        if len(types) == 0:
            raise InvalidStepsException(f"The step '{snode.id()}' has no inputs, and cannot be added to this workflow")

        if len(types) == 1:
            t = types[0]
            if len(input_parts) != 2:
                s = "/".join(input_parts)
                Logger.log(f"The node '{snode.id()}' did not correctly reference an input of the tool "
                           f"'{snode.step.tool().id()}', this was automatically corrected "
                           f"({s} → {lbl}/{t[0]})", LogLevel.WARNING)
            input_parts = [lbl, t[0]]
            return input_parts, t[1]
        elif len(input_parts) < 2:
            if guess_type is not None:
                # try to guess it from the outputs
                compatible_types = [ins[x] for x in ins if (not ins[x].input_type.optional) and ins[x].input_type.can_receive_from(guess_type)]
                if len(compatible_types) == 1:
                    inp = compatible_types[0]
                    Logger.info(f"Guessed the compatible match for '{node.id()}' with source type '{guess_type.id()}'"
                                f" → '{inp.input_type.id()}' ('{node.id()}' → '{node.id()}/{inp.tag}')")
                    return [lbl, inp.tag], inp.input_type
                else:
                    # Should this step also take into consideration the _optional_ nature of the node,
                    # ie: should it favour required nodes if that's an option
                    ultra_compatible = [ins[x] for x in ins if (not ins[x].input_type.optional) and type(ins[x].input_type) == type(guess_type)]

                    if len(ultra_compatible) == 1:
                        ultra = ultra_compatible[0]
                        Logger.warn(f"There were {len(compatible_types)} matched types for the node '{node.id()}', "
                                    f"the program has guessed an exact compatible match of "
                                    f"type '{ultra.input_type.id()}' to tag '{ultra.tag}'")
                        return [lbl, ultra.tag], ultra.input_type
                    else:
                        s = "/".join(input_parts)
                        compat_str = ", ".join(f"{x.tag}: {x.input_type.id()}" for x in compatible_types)
                        raise Exception(f"The end node '{node.id()}' did not specify where to join '{guess_type.id()}'"
                                        f" to, there were {len(compatible_types)} compatible types"
                                        f" ({compat_str}) and {len(ultra_compatible)} exact types."
                                        f" You will need to provide more information to proceed.")
            else:

                possible_tags = ", ".join(f"'{x}'" for x in ins)
                s = "/".join(input_parts)
                Logger.critical(f"The tag '{s}' could not uniquely identify an input of '{snode.id()}', requires the "
                                f"one of the following tags: {possible_tags}")
                return input_parts, None

        else:
            tag = input_parts[1]
            tool_input: ToolInput = snode.step.tool().inputs_map().get(tag)
            if tool_input:
                if len(input_parts) != 2:
                    s = "/".join(input_parts)
                    Logger.log(f"The node '{snode.id()}' did not correctly reference an input of the tool "
                               f"'{snode.step.tool().id()}', this was automatically corrected "
                               f"({s} → {lbl}/{tag})", LogLevel.WARNING)
                input_parts = [lbl, tag]
                return input_parts, tool_input.input_type

            possible_tags = ", ".join(f"'{x}'" for x in ins)
            raise Exception(f"Could not identify an input called '{tag}' on the node '{snode.id()}' "
                            f", possible tags: {possible_tags}")

        raise Exception("Unhandled pathway in 'get_tag_and_type_from_node'")

    @staticmethod
    def guess_connection_between_nodes(s_node: Node, f_node: Node) \
            -> Optional[Tuple[Tuple[List[str], DataType], Tuple[List[str], DataType]]]:
        outs, ins = s_node.outputs(), f_node.inputs()

        s_types: List[Tuple[List[str], DataType]] = [([s_node.id(), x], outs[x].output_type) for x in outs]
        f_types: List[Tuple[List[str], DataType]] = [([f_node.id(), x], ins[x].input_type) for x in ins if x.ou]

        # O(n**2) for determining types
        matching_types: List[Tuple[Tuple[List[str], DataType], Tuple[List[str], DataType]]] = []
        for s_type in s_types:
            matching_types.extend([(s_type, f_type) for f_type in f_types if f_type[1].can_receive_from(s_type[1])])

        if len(matching_types) == 0:
            raise InvalidInputsException(f"Can't find a common type between '{s_node.id()}' and '{f_node.id()}'")

        if len(matching_types) > 1:
            compat = [str(t) for t in matching_types]
            message = f"Couldn't guess the single connection between '{s_node.id()}' and '{f_node.id()}', " \
                f"there was {len(compat)} compatible connections ({compat})"
            Logger.log(message, LogLevel.CRITICAL)
            return None

        matched = matching_types[0]
        Logger.info(f"Guessed the connection between nodes '{s_node.id()}")
        return matched

    # TRANSLATIONS

    def dump_translation(self, translation: SupportedTranslation, to_console=True, to_disk=False, with_docker=True, with_hints=False,
                         with_resource_overrides=False):
        return translations.dump_translation(self,
                                             translation=translation,
                                             to_console=to_console,
                                             to_disk=to_disk,
                                             with_docker=with_docker,
                                             with_hints=with_hints,
                                             with_resource_overrides=with_resource_overrides)

    # def cwl(self, is_nested_tool=False, with_docker=True, with_hints=False, with_resource_overrides=False) -> Tuple[cwl.Workflow, dict, List[cwl.Serializable]]:
    #
    #     metadata = self.metadata() if self.metadata() else WorkflowMetadata()
    #     w = cwl.Workflow(self.identifier, self.friendly_name(), metadata.documentation)
    #
    #     w.inputs: List[cwl.InputParameter] = [i.cwl() for i in self._inputs]
    #
    #     if with_resource_overrides:
    #         rOverride = cwl.InputParameter(
    #             self.RESOURCE_OVERRIDE_KEY,
    #             param_type=["null", self._generate_cwl_resource_override_schema_for_steps()]
    #         )
    #         w.inputs.append(rOverride)
    #
    #     w.steps: List[cwl.WorkflowStep] = [s.cwl(is_nested_tool=is_nested_tool) for s in self._steps]
    #     w.outputs = [o.cwl() for o in self._outputs]
    #
    #     #
    #     keys = ["coresMin", "coresMax", "ramMin", "ramMax"]
    #     sins = [ToolInput(k, Int(optional=True)).cwl() for k in keys]
    #     if with_resource_overrides:
    #         for s in w.steps:
    #             # work out whether (the tool of) s is a workflow or tool
    #             resource_override_step_inputs = [cwl.WorkflowStepInput(
    #                 input_id=k,
    #                 source=self.RESOURCE_OVERRIDE_KEY,
    #                 value_from=f"${{var k = \"{k}\";var stepId = \"{s.id}\";if(!self) return null;if (!(stepId in self)) return null;return self[stepId][k]}}"
    #                 # value_from=f"$(!self ? null : (!self['{s.id}'] ? null : self['{s.id}']['{k}']))"
    #             ) for k in keys]
    #             s.inputs.extend(resource_override_step_inputs)
    #
    #     if with_hints:
    #         resource_schema = get_cwl_schema_for_recognised_hints()
    #         nullable_resource_schema = ["null", resource_schema]
    #         w.inputs.append(cwl.InputParameter(self.HINTS_KEY, param_type=nullable_resource_schema))
    #         for s in w.steps:
    #             s.inputs.append(cwl.WorkflowStepInput(self.HINTS_KEY, self.HINTS_KEY))
    #
    #     w.requirements.append(cwl.InlineJavascriptReq())
    #     w.requirements.append(cwl.StepInputExpressionRequirement())
    #
    #     if self.has_scatter:
    #         w.requirements.append(cwl.ScatterFeatureRequirement())
    #     if self.has_subworkflow:
    #         w.requirements.append(cwl.SubworkflowFeatureRequirement())
    #     if self.has_multiple_inputs:
    #         w.requirements.append(cwl.MultipleInputFeatureRequirement())
    #
    #     tools = []
    #     tools_to_build: Dict[str, Tool] = {s.step.tool().id(): s.step.tool() for s in self._steps}
    #     for t in tools_to_build:
    #         tool: Tool = tools_to_build[t]
    #         if isinstance(tool, Workflow):
    #             wf_cwl, _, subtools = tool.cwl(is_nested_tool=True, with_docker=with_docker, with_hints=with_hints, with_resource_overrides=with_resource_overrides)
    #             tools.append(wf_cwl)
    #             tools.extend(subtools)
    #         elif isinstance(tool, CommandTool):
    #             tool_cwl = tool.cwl(with_docker=with_docker)
    #             if with_hints:
    #                 tool_cwl.inputs.append(
    #                     cwl.InputParameter(self.HINTS_KEY, param_type=["null", get_cwl_schema_for_recognised_hints()]))
    #                 hm = tool.hint_map()
    #                 if hm:
    #                     tool_cwl.requirements.append(cwl.ResourceRequirement(
    #                         cores_min=self._generate_hint_selectors_for_hint_map("coresMin", hm),
    #                         cores_max=self._generate_hint_selectors_for_hint_map("coresMax", hm),
    #                         ram_min=self._generate_hint_selectors_for_hint_map("ramMin", hm),
    #                         ram_max=self._generate_hint_selectors_for_hint_map("ramMax", hm)
    #                     ))
    #
    #             if with_resource_overrides:
    #                 tool_cwl.inputs.extend(sins)
    #
    #             tools.append(tool_cwl)
    #         else:
    #             raise Exception(f"Unknown tool type: '{type(tool)}'")
    #
    #     inp = {i.id(): i.input.cwl_input() for i in self._inputs}
    #
    #     return w, inp, tools

    #     @staticmethod
    #     def _generate_cwl_resource_override_schema():
    #         schema = cwl.CommandInputRecordSchema()
    #         schema.fields = [
    #             cwl.CommandInputRecordSchema.CommandInputRecordField("coresMin", ["long", "string", "null"]),
    #             cwl.CommandInputRecordSchema.CommandInputRecordField("coresMax", ["int", "string", "null"]),
    #             cwl.CommandInputRecordSchema.CommandInputRecordField("ramMin", ["long", "string", "null"]),
    #             cwl.CommandInputRecordSchema.CommandInputRecordField("ramMax", ["int", "string", "null"])
    #         ]
    #         return schema
    #
    #     def _generate_cwl_resource_override_schema_for_steps(self):
    #         schema = cwl.CommandInputRecordSchema()
    #
    #         for step in self._steps:
    #             tool = step.step.tool()
    #             if isinstance(tool, Workflow):
    #                 key = tool.id() + "_resource_override" # self.RESOURCE_OVERRIDE_KEY
    #                 override_schema = tool._generate_cwl_resource_override_schema_for_steps()
    #             else:
    #                 key = step.step.id()
    #                 override_schema = self._generate_cwl_resource_override_schema()
    #
    #             schema.fields.append(cwl.CommandInputRecordSchema.CommandInputRecordField(key, ["null", override_schema]))
    #
    #         return schema
    #
    #     @staticmethod
    #     def _generate_hint_selectors_for_hint_map(resource_key, hint_map):
    #         import json
    #         return f"""${{
    #     var key = '{resource_key}';
    #     if (inputs["{Workflow.RESOURCE_OVERRIDE_KEY}"] && inputs["{Workflow.RESOURCE_OVERRIDE_KEY}"][key])
    #         return inputs["{Workflow.RESOURCE_OVERRIDE_KEY}"][key];
    #     var hints = inputs.hints;
    #     if (!hints) return null;
    #
    #     var hintMap = {json.dumps(hint_map)};
    #     for (var hint in hintMap) {{
    #         var providedHintValue = hints[hint];
    #         if (!providedHintValue || !(providedHintValue in hintMap[hint])) continue;
    #         var hintValueToResourceMap = hintMap[hint][providedHintValue];
    #         if (hintValueToResourceMap[key]) return hintValueToResourceMap[key];
    #     }}
    #     return null;
    # }}"""# .replace("    ", "").replace("\n", "")

    @staticmethod
    def build_aliases(steps):
        """

        :param steps:
        :return:
        """

        get_alias = lambda t: t[0] + "".join([c for c in t[1:] if c.isupper()])
        aliases = set()

        tools: List[Tool] = [s.step.tool() for s in steps]
        tool_name_to_tool: Dict[str, Tool] = {t.id().lower(): t for t in tools}
        tool_name_to_alias = {}
        steps_to_alias: Dict[str, str] = {s.id().lower(): get_alias(s.id()).lower() for s in steps}

        for tool in tool_name_to_tool:
            a = get_alias(tool).upper()
            s = a
            idx = 2
            while s in aliases:
                s = a + str(idx)
                idx += 1
            aliases.add(s)
            tool_name_to_alias[tool] = s

        return tool_name_to_alias, steps_to_alias

    def wdl(self, with_docker=True, is_nested_tool=False):
        """

        :param with_docker:
        :param is_nested_tool:
        :return:
        """

        # Notes:
        #   All classes have a .wdl() property that returns the wdl classes,
        #       EXCEPT Workflow which returns (wdl.Workflow, string, (wdl.Task | wdl.Workflow)[])
        #       All wdlgen classes have a .get_string(**kwargs) function
        #       The wdlgen Workflow class requires a

        w = wdl.Workflow(self.identifier)
        tools: List[Tool] = [s.step.tool() for s in self._steps]

        wtools = {}  # Store all the tools by their name in this dictionary
        tool_aliases, step_aliases = self.build_aliases(self._steps)  # Generate call and import aliases

        # Convert self._inputs -> wdl.Input
        for i in self._inputs:
            wd = i.input.data_type.wdl()
            w.inputs.append(wdl.Input(wd, i.id(), i.input.data_type.default()))

        # Convert self._outputs -> wdl.Output
        w.outputs = [
            wdl.Output(
                o.output.data_type.wdl(),
                o.id(),
                "{a}.{b}".format(  # Generate fully qualified stepid.tag identifier (MUST be step node)
                    a=first_value(first_value(o.connection_map).source_map).start.id(),
                    b=first_value(first_value(o.connection_map).source_map).stag
                ))
            for o in self._outputs]

        # Generate import statements (relative tool dir is later?)
        w.imports = [
            wdl.Workflow.WorkflowImport(
                t.id(),
                tool_aliases[t.id().lower()].upper(),
                None if is_nested_tool else "tools/")
            for t in tools]

        # Step[] -> (wdl.Task | wdl.Workflow)[]
        for s in self._steps:
            t = s.step.tool()

            if isinstance(t, Workflow):
                wf_wdl, _, wf_tools = t.wdl(with_docker=with_docker, is_nested_tool=True)
                wtools[s.id()] = wf_wdl
                wtools.update(wf_tools)
            else:
                wtools[t.id()] = t.wdl(with_docker=with_docker)

            w.calls.append(
                s.wdl(tool_aliases[t.id().lower()].upper() + "." + t.id(), s.id())
            )

        inp = {f"{self.identifier}.{i.id()}": i.input.wdl_input() for i in self._inputs}

        return w, inp, wtools

    def get_cwl_dicts(self, with_docker=True):
        cwl_data, inp_data, tools_ar = self.cwl(with_docker=with_docker)
        return cwl_data.get_dict(), inp_data, [t.get_dict() for t in tools_ar]

    # def dump_cwl(self, to_disk: False, with_docker=True, write_inputs_file=True):
    #     import os, yaml
    #     cwl_data, inp_data, tools_ar = self.get_cwl_dicts(with_docker=with_docker)
    #
    #     d = os.path.expanduser("~") + f"/Desktop/{self.identifier}/cwl/"
    #     d_tools = d + "tools/"
    #
    #     if not os.path.isdir(d):
    #         os.makedirs(d)
    #     if not os.path.isdir(d_tools):
    #         os.makedirs(d_tools)
    #
    #     from cwlgen.cwlgen.utils import literal, literal_presenter
    #     yaml.add_representer(literal, literal_presenter)
    #
    #     print(yaml.dump(cwl_data, default_flow_style=False))
    #     print(yaml.dump(inp_data, default_flow_style=False))
    #     for t in tools_ar:
    #         print(yaml.dump(t, default_flow_style=False))
    #
    #     if to_disk:
    #         os.chdir(d)
    #         wf_filename = d + self.id() + ".cwl"
    #         with open(wf_filename, "w+") as cwl:
    #             Logger.log(f"Writing {self.identifier}.cwl to disk")
    #             yaml.dump(cwl_data, cwl, default_flow_style=False)
    #             Logger.log(f"Written {self.identifier}.cwl to disk")
    #
    #         if write_inputs_file:
    #             with open(d + self.id() + "-job.yml", "w+") as cwl:
    #                 Logger.log(f"Writing {self.identifier}-job.yml to disk")
    #                 yaml.dump(inp_data, cwl, default_flow_style=False)
    #                 Logger.log(f"Written {self.identifier}-job.yml to disk")
    #         else:
    #             Logger.log("Skipping writing input (yaml) job file")
    #
    #         # z = zipfile.ZipFile(d + "tools.zip", "w")
    #         for tool in tools_ar:
    #             tool_name = tool["id"]
    #             tool_filename = tool_name + ".cwl"
    #             with open(d_tools + tool_filename, "w+") as cwl:
    #                 Logger.log(f"Writing {tool_name}.cwl to disk")
    #                 yaml.dump(tool, cwl, default_flow_style=False)
    #                 Logger.log(f"Written {tool_name}.cwl to disk")
    #
    #         import subprocess
    #
    #         Logger.info("Validing outputted CWL")
    #
    #         cwltool_result = subprocess.run(["cwltool", "--validate", wf_filename])
    #         if cwltool_result.returncode == 0:
    #             Logger.info("Exported workflow is valid CWL.")
    #         else:
    #             Logger.critical(cwltool_result.stderr)
    #
    #         Logger.info("Zipping tools")
    #         os.chdir(d)
    #
    #         zip_result = subprocess.run(["zip", "-r", "tools.zip", "tools/"])
    #         if zip_result.returncode == 0:
    #             Logger.info("Zipped tools")
    #         else:
    #             Logger.critical(zip_result.stderr)
    #
    #         #     z.write("tools/" + tool_filename)
    #         # z.close()

    def dump_wdl(self, to_disk: False, with_docker=True):
        import os, json
        wf, inp, tls = self.wdl(with_docker=with_docker)

        wfd = wf.get_string()
        wftools = {t: tls[t].get_string() for t in tls}

        print(wfd)
        print("================")
        print(json.dumps(inp))
        print("================")
        print("\n*******\n".join(wftools.values()))

        d = os.path.expanduser("~") + f"/Desktop/{self.identifier}/wdl/"
        d_tools = d + "tools/"
        if not os.path.isdir(d):
            os.makedirs(d)
        if not os.path.isdir(d_tools):
            os.makedirs(d_tools)

        if to_disk:
            with open(d + self.identifier + ".wdl", "w+") as wdl:
                Logger.log(f"Writing {self.identifier}.get_string to disk")
                wdl.write(wfd)
                Logger.log(f"Written {self.identifier}.get_string to disk")

            with open(d + self.identifier + "-job.json", "w+") as inpFile:
                Logger.log(f"Writing {self.identifier}-job.json to disk")
                json.dump(inp, inpFile)
                Logger.log(f"Written {self.identifier}-job.json to disk")

            for tool_name in wftools:
                tool = wftools[tool_name]
                with open(d_tools + tool_name + ".wdl", "w+") as wdl:
                    Logger.log(f"Writing {tool_name}.get_string to disk")
                    wdl.write(tool)
                    Logger.log(f"Written {tool_name}.get_string to disk")



from typing import List, Dict, Any, Optional

# import cwlgen.cwlgen as cwl
from janis.graph.node import Node, NodeTypes
from janis.graph.stepinput import StepInput
from janis.tool.tool import Tool, ToolInput, ToolOutput
from janis.utils.logger import Logger


class Step:
    def __init__(self, identifier: str, tool: Tool, meta: Optional[Any]=None,
                 label: str = None, doc: str = None):
        self._identifier: str = identifier.lower()
        self.label = label
        self.doc = doc

        self.__tool: Tool = tool
        self.__meta: Optional[Any] = meta

    def id(self):
        return self._identifier

    def __str__(self):
        t = self.__tool.id()
        return f"{self.id()}: {t}"

    def __repr__(self):
        return str(self)

    def set_input_value(self, tag: str, value: str):
        l = self._identifier
        Logger.log(f"Updating '{l}': setting '{tag}' -> '{value}'")
        self.__meta[tag] = value

    def requires(self) -> Dict[str, ToolInput]:
        return self.tool().inputs_map()

    def provides(self) -> Dict[str, ToolOutput]:
        return self.tool().outputs_map()

    def tool(self) -> Tool:
        # haha don't mispell this, otherwise infinite recursion through __getattr__
        return self.__tool

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]

        if item in self.tool().inputs_map():

            return f"{self.id()}/{self.tool().inputs_map()[item].tag}", self
        if item in self.tool().outputs_map():
            return f"{self.id()}/{self.tool().outputs_map()[item].tag}", self

        tags = ", ".join([f"in.{i.tag}" for i in self.tool().inputs()]
                         + [f"out.{o.tag}" for o in self.tool().outputs()])

        raise AttributeError(f"Step '{self.id()}' with tool '{self.tool().id()}' has no identifier '{item}' ({tags})")


class StepNode(Node):
    def __init__(self, step: Step):
        super().__init__(NodeTypes.TASK, step.id())
        self.step = step

    def inputs(self) -> Dict[str, ToolInput]:
        return self.step.requires()

    def outputs(self) -> Dict[str, ToolOutput]:
        return self.step.provides()

    def __getattr__(self, item):
        if item in self.__dict__:
            return self.__dict__[item]

        if item in self.inputs() or item in self.outputs():
            return f"{self.id()}/{item}", self

        raise AttributeError(f"type object '{type(self)}' has no attribute '{item}'")

    def __repr__(self):
        return f"{self.node_type}: {self.step.tool().id()}"

    # def cwl(self, is_nested_tool=False):
    #
    #     step = self.step.cwl(is_nested_tool)
    #
    #     ins = self.inputs()
    #     scatterable = []
    #
    #     for k in ins:
    #         inp = ins[k]
    #         if k not in self.connection_map:
    #             if inp.input_type.optional:
    #                 continue
    #             else:
    #                 raise Exception(f"Error when building connections for step '{self.id()}', "
    #                                 f"could not find required connection: '{k}'")
    #
    #         inp_t = self.inputs()[k].input_type
    #         edge = self.connection_map[k]
    #         default = edge.default if edge.default else inp_t.default()
    #         d = cwl.WorkflowStepInput(
    #             input_id=inp.tag,
    #             source=edge.source(),
    #             link_merge=None,        # this will need to change when edges have multiple source_map
    #             default=default,
    #             value_from=None
    #         )
    #         if edge.has_scatter():
    #             scatterable.append(k)
    #
    #         step.inputs.append(d)
    #
    #     if len(scatterable) > 0:
    #         if len(scatterable) > 1:
    #             Logger.info("Discovered more than one scatterable field on step '{step_id}', "
    #                         "deciding scatterMethod to be dot_product".format(step_id=self.id()))
    #             step.scatterMethod = "dot_product"
    #         step.scatter = scatterable
    #     return step

    def wdl_map(self) -> List[str]:
        q = []
        for inp in self.step.tool().inputs():
            if inp.tag in self.connection_map:
                si: StepInput = self.connection_map[inp.tag]
                q.append("{tag} = {value}".format(tag=si.ftag, value=si.dotted_source()))
            elif not inp.optional:
                raise Exception(f"Required option '{inp.tag}' for step '{self.id()}' "
                                f"was not found during conversion to WDL")
        return q

    def wdl(self, step_identifier: str, step_alias: str):
        import wdlgen.wdlgen as wdl

        ins = self.inputs()

        # One step => One WorkflowCall. We need to traverse the edge list to see if there's a scatter
        # then we can build up the WorkflowCall / ScatterCall
        scatterable = [self.connection_map[k].dotted_source()
                       for k in self.inputs() if k in self.connection_map and self.connection_map[k].has_scatter()]

        # We need to replace the scatterable key(s) with some random variable, eg: for i in iterable:
        ordered_variable_identifiers = ["i", "j", "k", "x", "y", "z", "a", "b", "c", "ii", "jj", "kk", "xx", "yy", "zz"]
        new_to_old_identifier = {k.dotted_source(): k.dotted_source() for k in self.connection_map.values()
                                 if not isinstance(k.dotted_source(), list)}

        # We'll wrap everything in the scatter block later, but let's replace the fields we need to scatter
        # with the new scatter variable (we'll try to guess one based on the fieldname. We might need to eventually
        # pass the workflow inputs to make sure now conflict will arise.
        # Todo: Pass Workflow input tags to wdl scatter generation to ensure scatter var doesn't conflict with inputs
        for s in scatterable:
            new_var = s.split(".")[-1][0].lower()
            while new_var in new_to_old_identifier:
                new_var = ordered_variable_identifiers.pop(0)
            new_to_old_identifier[s] = new_var

        # Let's map the inputs, to the source.
        # We're using a dictionary for the map atm, but WDL requires the format:
        #       fieldName: sourceCall.Output

        inputs_map = {}
        for k in ins:
            inp = ins[k]
            if k not in self.connection_map:
                if inp.input_type.optional:
                    continue
                else:
                    raise Exception(f"Error when building connections for step '{self.id()}', "
                                    f"could not find required connection: '{k}'")

            edge = self.connection_map[k]
            ds = edge.dotted_source()

            if isinstance(ds, list):
                if len(ds) == 1:
                    ds = ds[0]
                elif len(ds) > 1:
                    Logger.critical("Conversion to WDL does not currently support multiple sources")
                    ds = f'[{", ".join(ds)}]'

            if ds in new_to_old_identifier and new_to_old_identifier[ds]:
                inputs_map[k] = new_to_old_identifier[ds]
            elif edge.default is not None:
                if isinstance(edge.default, bool):
                    inputs_map[k] = "true" if edge.default else "false"
                elif isinstance(edge.default, str):
                    inputs_map[k] = f'"{edge.default}"'
                else:
                    inputs_map[k] = edge.default
            else:
                inputs_map[k] = ds

        call = wdl.WorkflowCall(step_identifier, step_alias, inputs_map)

        for s in scatterable:
            call = wdl.WorkflowScatter(new_to_old_identifier[s], s, [call])

        return call




#
# Compile a java file
#

from typing import List

from janis.tool.commandtool import ToolInput, ToolOutput, ToolArgument
from janis.types.common_data_types import File
from janis.unix.tools.unixtool import UnixTool


class Compile(UnixTool):

    @staticmethod
    def tool():
        return "javaCompiler"

    @staticmethod
    def base_command():
        return "javac"

    def friendly_name(self):
        return "Java compiler"

    @staticmethod
    def docker():
        return "openjdk:8"

    def arguments(self) -> List[ToolArgument]:
        return [ToolArgument(".", "-d")]    # CurrentWorkingDirectory()

    def inputs(self):
        return [
            ToolInput("file", File(), position=1)
        ]

    def outputs(self):
        return [
            ToolOutput("compiled", File(), glob="*.class")
        ]





from typing import Dict, Optional, Any

from janis.graph.node import Node, NodeTypes
from janis.types.data_types import DataType
from janis.tool.commandtool import ToolOutput, ToolInput


class Output:
    """
        Only catch with output is we infer the type, we don't explicitly define it
    """
    def __init__(self, identifier: str, meta: Any = None,
                 label: str=None, doc: str=None):
        self._identifier: str = identifier
        self.label = label
        self.doc = doc

        self.data_type: Optional[DataType] = None
        self.meta = meta

    def id(self) -> str:
        return self._identifier


class OutputNode(Node):

    def __init__(self, output: Output):
        self.output = output
        super().__init__(NodeTypes.OUTPUT, output.id())

    def inputs(self) -> Dict[str, ToolInput]:
        return {self.output._identifier: ToolInput("outp", self.output.data_type)}

    def outputs(self) -> Dict[str, ToolOutput]:
        return {}



from janis import Array
from janis.tool.commandtool import ToolInput, ToolOutput
from janis.types.common_data_types import File
from janis.unix.data_types.tar_file import TarFile
from janis.unix.tools.unixtool import UnixTool


class Untar(UnixTool):

    @staticmethod
    def base_command():
        return ["tar", "xf"]

    def friendly_name(self):
        return "Untar archive"

    @staticmethod
    def tool():
        return "Untar"

    def inputs(self):
        return [
            ToolInput("tarFile", TarFile(), position=0)
        ]

    def outputs(self):
        return [
            ToolOutput("files", Array(File()), glob="*.java")
        ]

"""

EXAMPLE XML

<Step>
    <title>Untar</title>
    <author>Michael Franklin</author>
    <cores>1</cores>
    <ram>1024</ram>
    <label>{{label}}</label>
    <supportedTypes>
        <type>cwl</type>
        <type>wdl</type>
    </supportedTypes>
    
    <inputs>
        <input>
            <tag>input</tag>
            <type>File</type>
        </input>
    </inputs>
    <outputs>
        <output>
            <tag>output</tag>
            <type>File</type>
        </output>
    </outputs>
</Step>

+ provide CWL template

#!/usr/bin/env cwl-runner

cwlVersion: v1.0
class: CommandLineTool
baseCommand: [tar, xf]
inputs:
  input:
    type: File
    inputBinding:
      position: 1
  output:
    type: string
    inputBinding:
      position: 2

outputs:
  example_out:
    type: File
    outputBinding:
      glob: $(inputs.outputfile)

+ wdl template

task {
    File input
    File output
    command {
        tar -xf ${input}
    }
}

"""


from setuptools import setup, find_packages
from janis import __version__
from janis.utils.janisconstants import GITHUB_URL

DESCRIPTION = "Contains classes and helpers to build a workflow, and provide options to convert to CWL / WDL"

######## SHOULDN'T NEED EDITS BELOW THIS LINE ########

with open("./janis/README.md") as readme:
    long_description = readme.read()

setup(
    name="janis pipelines",
    version=__version__,
    description=DESCRIPTION,
    url=GITHUB_URL,
    author="Michael Franklin, Evan Thomas, Mohammad Bhuyan",
    author_email="michael.franklin@petermac.org",
    license="GNU",
    packages=["janis"] + ["janis." + p for p in sorted(find_packages('./janis'))],
    install_requires=["networkx>=2.1", "ruamel.yaml >= 0.12.4, <= 0.15.77", "six", "illusional.wdlgen",
                      "illusional.cwlgen>=0.0.4"],
    zip_safe=False,
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Bio-Informatics'
    ],
    extras_require={
        'bioinformatics': ['janis.bioinformatics']
    }
)



from janis.utils.janisconstants import GITHUB_URL
from .cwl import dump_cwl as cwl_dump_workflow

SupportedTranslation = str


class SupportedTranslations:
    CWL: SupportedTranslation = "cwl"
    WDL: SupportedTranslation = "wdl"


def dump_translation(workflow, translation: SupportedTranslation, to_console=True, to_disk=False, with_docker=True,
                     with_hints=False, with_resource_overrides=False):
    if translation == SupportedTranslations.CWL:
        return cwl_dump_workflow(workflow, to_console=to_console, to_disk=to_disk, with_docker=with_docker,
                                 with_hints=with_hints, with_resource_overrides=with_resource_overrides)

    elif translation == SupportedTranslations.WDL:
        raise NotImplementedError("WDL has not been implemented yet, please raise an issue on Github: " + GITHUB_URL)

    else:
        raise NotImplementedError(f"The requested translation ('{translation}') has not been implemented yet, "
                                  f"prepare it yourself at '{GITHUB_URL}'.")



###################
# Implementations #
###################
from typing import Dict, Any
import wdlgen as wdl
import cwlgen as cwl

from janis.types.data_types import DataType, NativeTypes, NativeType


class String(DataType):

    @staticmethod
    def name():
        return "String"

    @staticmethod
    def primitive():
        return NativeTypes.kStr

    def doc(self):
        return "A string"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "string", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))

    def can_receive_from(self, other):
        if isinstance(other, Filename):
            return True
        return super().can_receive_from(other)


class Filename(String):

    def __init__(self, extension: str=None):
        """
        :param extension: with no '.' (dot)
        """
        self.extension = extension
        super().__init__(optional=True, default=self.generated_filename())

    @staticmethod
    def name() -> str:
        return "Filename"

    @staticmethod
    def primitive() -> NativeType:
        return NativeTypes.kStr

    def cwl_type(self):
        self.optional = False
        t = super().cwl_type()
        self.optional = True
        return t

    def doc(self) -> str:
        return """
This class is a placeholder for generated filenames, by default it is optional and CAN be overrided, 
however the program has been structured in a way such that these names will be generated based on the step label. 
These should only be used when the tool _requires_ a filename to output and you aren't 
concerned what the filename should be. The Filename DataType should NOT be used as an output.
""".strip()

    @classmethod
    def schema(cls) -> Dict:
        pass

    def map_cwl_type(self, parameter: cwl.Parameter):
        super().map_cwl_type(parameter)
        parameter.default = self.generated_filename()

    def generated_filename(self, prefix: str=None) -> str:
        import uuid
        pre = (prefix + "-") if prefix is not None else ""
        ex = "" if self.extension is None else self.extension
        return pre + "generated-" + str(uuid.uuid1()) + ex

    def can_receive_from(self, other: DataType):
        # Specific override because Filename should be able to receive from string
        if isinstance(other, String):
            return True  # Always provides default, and is always optional
        return super().can_receive_from(other)


class Int(DataType):

    @staticmethod
    def name():
        return "Integer"

    @staticmethod
    def primitive():
        return NativeTypes.kInt

    def doc(self):
        return "An integer"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))


class Float(DataType):

    @staticmethod
    def name():
        return "Float"

    @staticmethod
    def primitive():
        return NativeTypes.kFloat

    def doc(self):
        return "A float"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))


class Double(DataType):

    @staticmethod
    def name():
        return "Double"

    @staticmethod
    def primitive():
        return NativeTypes.kDouble

    def doc(self):
        return "An integer"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "number", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))


class Boolean(DataType):

    @staticmethod
    def name():
        return "Boolean"

    @staticmethod
    def primitive():
        return NativeTypes.kBool

    def doc(self):
        return "A boolean"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "boolean", "required": True}

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))


class File(DataType):

    @staticmethod
    def name():
        return "File"

    @staticmethod
    def primitive():
        return NativeTypes.kFile

    def doc(self):
        return "A local file"

    @classmethod
    def schema(cls) -> Dict:
        return {
            "path": {"type": "string", "required": True}
        }

    def get_value_from_meta(self, meta):
        return meta.get("path")

    @staticmethod
    def cwl_input(value: Any):
        return {
            "class": cwl.CwlTypes.FILE,
            "path": value
        }

    # def wdl(self):
    #     # Todo: SECONDARY FILES
    #     return wdlgen.WdlType.parse_type(NativeTypes.map_to_wdl(self.primitive()) + self._question_mark_if_optional())


class Directory(DataType):
    @staticmethod
    def name():
        return "Directory"

    @staticmethod
    def primitive():
        return NativeTypes.kDirectory

    def doc(self):
        return "A directory of files"

    def get_value_from_meta(self, meta):
        return meta["path"]

    @classmethod
    def schema(cls) -> Dict:
        return {
            "path": {"type": "string", "required": True}
        }

    def input_field_from_input(self, meta):
        return meta["path"]

    @staticmethod
    def cwl_input(value: Any):
        # WDL: "{workflowName}.label" = meta["path"}
        return {
            "class": cwl.CwlTypes.DIRECTORY,
            "path": value
        }


class Array(DataType):

    def __init__(self, t: DataType, optional=False):
        if not isinstance(t, DataType):
            raise Exception(f"Type t ({type(t)}) must be an instance of 'Type'")

        self._t = t
        super().__init__(optional)

    def subtype(self):
        return self._t

    @staticmethod
    def name():
        return "Array"

    @staticmethod
    def primitive():
        return NativeTypes.kArray

    def id(self):
        if self._t is None:
            return super().id()
        t = self._t
        typed = f"Array<{t.id()}>"
        if self.optional:
            return f"Optional<{typed}>"
        return typed

    def doc(self):
        return "An array"

    @classmethod
    def schema(cls) -> Dict:
        return {"type": "array"}

    def cwl_type(self):
        inp = cwl.CommandInputArraySchema(
            items=self._t.cwl_type(),
            # label=None,
            # input_binding=None
        )
        return [inp, "null"] if self.optional else inp

    def map_cwl_type(self, parameter: cwl.Parameter) -> cwl.Parameter:
        parameter.type = cwl.CommandInputArraySchema(
            items=None,
            label=None,
            input_binding=None
        )
        return parameter

    # def wdl(self):
    #     return f"{NativeTypes.map_to_wdl(self.primitive())}[{NativeTypes.map_to_wdl(self._t.primitive())}]"

    def wdl(self) -> wdl.ArrayType:
        return wdl.ArrayType(self._t.wdl(), requires_multiple=False)

    def can_receive_from(self, other):
        if isinstance(other, Array):
            return self._t.can_receive_from(other._t)
        if not self._t.can_receive_from(other):
            return False
        return super().can_receive_from(other)

    def input_field_from_input(self, meta):
        return next(iter(meta.values()))


class Stdout(File):

    @staticmethod
    def name():
        return "Stdout"

    @staticmethod
    def primitive():
        return NativeTypes.kStdout

    def __init__(self, subtype=File(), stdoutname=None):
        super().__init__(optional=False)
        self.subtype = subtype if subtype is not None else File()
        self.stdoutname = stdoutname

    def received_type(self):
        return self.subtype



from janis import File, Array, ToolInput, ToolOutput, Stdout
from janis.unix.tools.unixtool import UnixTool


class Merge(UnixTool):

    def friendly_name(self) -> str:
        return "Merge Files"

    @staticmethod
    def tool():
        return "merge"

    @staticmethod
    def base_command():
        return ["cat"]

    def inputs(self):
        return [
            ToolInput("files", Array(File()))
        ]

    def outputs(self):
        return [
            ToolOutput("merged", Stdout())
        ]



from typing import Dict


def first_value(d: Dict):
    return next(iter(d.values()))


def convert_expression_to_wdl(expression):
    import re

    # Some examples
    # $(inputs.filename) -> "{filename}"
    # $(inputs.filename).out -> "{filename}.out"
    # randomtext.filename -> "randomtext.filename

    if not expression:
        return expression

    if not isinstance(expression, str):
        raise Exception(f"Expected expression of type string, got: {expression} ({type(expression)})")

    r1 = re.compile(r"\$\(inputs\..*\)")
    m1 = r1.match(expression)
    if m1:
        return f'"{{{m1.group()[9:-1]}}}{expression[m1.span()[1]:]}"'

    r2 = re.compile(r"\$\(\..*\)")
    m2 = r2.match(expression)
    if m2:
        return f'"{{{m2.group()[9:-1]}}}{expression[m1.span()[1]:]}"'

    return f'"{expression}"'



# from typing import Dict
#
# import json
#
# # from pipeline_definition.types.input_type import InputType
# from Workflow.step import Step
#
#
# class StepContext:
#   def __init__(self, step):
#     self.__step = step
#     self.__branch_outputs_stack = list()
#     self.__dependency_contexts = list()
#
#   def inherit_context_of_branch(self, prev_ctx):
#     outputs_stack_from_prev_context = prev_ctx.output_stack_of_branch()
#     if not outputs_stack_from_prev_context:
#       return
#     self.__branch_outputs_stack.extend(outputs_stack_from_prev_context)
#
#   def output_stack_of_branch(self):
#     output_stack = list()
#
#     provides_dict = {self.__step.id(): self.__step.provides()}
#
#     output_dict = {self.__step.tag(): provides_dict}
#     output_stack.append(output_dict)
#
#     output_stack.extend(self.__branch_outputs_stack)
#
#     return output_stack
#
#   def add_dependency_context_from(self, step_ctx):
#     dependency_context = step_ctx.provides()
#     self.__dependency_contexts.append(dependency_context)
#     # Dependency search is not only limited to last step in referred branch
#     self.__dependency_contexts.extend(step_ctx.output_stack_of_branch())
#
#   def provides(self):
#     provides_dict = {self.__step.id(): self.__step.provides()}
#     output_dict = {self.__step.tag(): provides_dict}
#     return output_dict
#
#   def provides_for(self, step):
#     provides = self.provides()
#
#     if not provides:
#       return []
#
#     tag_specific = provides.get(step.tag())
#
#     if not tag_specific:
#       return []
#
#     step_specific = tag_specific.get(step.id())
#     if not step_specific:
#       return []
#
#     return step_specific
#
#   def to_json(self):
#     return json.dumps(self, default=lambda o: o.__dict__, sort_keys=True, indent=4)
#
#   def print(self):
#     print(self.to_json())
#
#   def map_input(self, step_input):
#
#     # mapping provided?
#     provided_mapping = self.__step.provided_value_for_requirement(step_input[Step.STR_ID])
#     if provided_mapping:
#       doc = {'provided': provided_mapping}
#     else:
#       doc = {'provided': ""}
#
#     if provided_mapping:
#       dependency_spec = Step.dependency_spec_from(provided_mapping)
#       candidates = self.find_match_for_input_dependency(step_input, dependency_spec)
#     else:
#       candidates = self.find_match_for_input(step_input)
#
#     if not candidates:
#       candidates['ERROR'] = "Failed to find an input candidate."
#
#     doc['candidates'] = candidates
#     return doc
#
#   def map_input_for_translation(self, step_input: InputType) -> Dict:
#
#     # mapping provided?
#     provided_mapping = self.__step.provided_value_for_requirement(step_input)
#
#     if provided_mapping:
#       dependency_spec = Step.dependency_spec_from(provided_mapping)
#       candidates = self.find_match_for_input_dependency(step_input, dependency_spec)
#     else:
#       candidates = self.find_match_for_input(step_input)
#
#     if not candidates:
#       raise Exception(f"Failed to find an input candidate for step: {self.__step.id()}, input requirement for: " +
#                       f"{step_input.type_name()}")
#
#     return candidates
#
#   def find_match_for_input_dependency(self, step_input, dependency_spec):
#
#     input_id = step_input[Step.STR_ID]
#     input_type = step_input[Step.STR_TYPE]
#     input_step_tag = self.__step.tag()
#
#     d_tag = dependency_spec['tag']
#
#     matches = {}
#     pref = 1
#
#     # Find the matching dependency in the stack
#     for dependency in self.__dependency_contexts:
#       matched = False
#       tag, step = next(iter(dependency.items()))
#       step_name, outputs = next(iter(step.items()))
#
#       if tag == d_tag:
#         # Found the matching dependency, now lets map the output
#
#         # Pass 1 to see if we have exact name / type match
#         for o in outputs:
#           o_id = o[Step.STR_ID]
#           o_type = o[Step.STR_TYPE]
#
#           # Name and Type match is heighest priority - conclusive
#           name = input_id
#           if name == o_id and input_type == o_type:
#             matches[pref] = self.__match_doc_for(o, step_name, input_step_tag)
#             pref = pref + 1
#             matched = True
#             break
#
#         if matched:
#           break
#
#         # Pass two is tag, tag_str type convention and type match
#         for o in outputs:
#           o_id = o[Step.STR_ID]
#           o_type = o[Step.STR_TYPE]
#
#           name = input_step_tag
#           if (o_id == name or o_id.startswith(name + "_")) and input_type == o_type:
#             matches[pref] = self.__match_doc_for(o, step_name, tag)
#             pref = pref + 1
#             matched = True
#             break
#
#         if matched:
#           break
#
#         # Pass three is type match
#         for o in outputs:
#           o_type = o[Step.STR_TYPE]
#
#           if input_type == o_type:
#             matches[pref] = self.__match_doc_for(o, step_name, tag)
#             pref = pref + 1
#
#         # Pass four is name match
#         for o in outputs:
#           o_id = o[Step.STR_ID]
#           if input_id == o_id:
#             matches[pref] = self.__match_doc_for(o, step_name, tag)
#             pref = pref + 1
#             continue
#
#     if matches:
#       return matches
#
#     return self.find_match_for_input(step_input)
#
#   def find_match_for_input(self, step_input: InputType):
#     # Finding the right input is based on the following criteria:
#     # 1. The output type of the nearest previous step matches
#     # 2. tags match                                        <<== This bit is not implemented yet
#     # 3. Search is limited if an input_scope is specified. <<== This bit is not implemented yet
#
#     step_tag = self.__step.tag()
#
#     matches = {}
#     pref = 1
#
#     # For each step in the stack, look at its provided outputs
#     for priority_entry in self.__branch_outputs_stack:
#       ostep_tag, ostep = next(iter(priority_entry.items()))
#       ostep_name, outputs = next(iter(ostep.items()))
#
#       for o_id, o_type in outputs.items():
#         if step_input == o_type: # and ostep_tag == step_tag:
#           matches[pref] = self.__match_doc_for(ostep_name, ostep_tag, o_id)
#           pref = pref + 1
#           break
#
#     return matches
#
#   @staticmethod
#   def __match_doc_for(name, tag, id):
#     doc = dict()
#     doc['step'] = name
#     doc['tag'] = tag
#     doc['id'] = id
#     return doc
